#!/usr/bin/env python
"""
riakasaurus trial test file.
riakasaurus _must_ be on your PYTHONPATH

"""

import json
import random
from twisted.trial import unittest
from twisted.python import log
from twisted.internet import defer

VERBOSE = False

from riakasaurus import riak

# uncomment to activate logging
# import sys
# log.startLogging(sys.stderr)

RIAK_CLIENT_ID = 'TEST'
BUCKET_PREFIX = 'riakasaurus.tests.'
JAVASCRIPT_SUM = """
function(v) {
  x = v.reduce(function(a,b){ return a + b }, 0);
  return [x];
}
"""


def randint():
    """Generate nice random int for our test."""
    return random.randint(1, 999999)


class Tests(unittest.TestCase):
    """
    trial unit tests.
    """

    test_keys = ['foo', 'foo1', 'foo2', 'foo3', 'bar', 'baz', 'ba_foo1',
                 'blue_foo1']

    @defer.inlineCallbacks
    def setUp(self):
        self.client = riak.RiakClient(client_id=RIAK_CLIENT_ID)
        self.bucket_name = BUCKET_PREFIX + self.id().rsplit('.', 1)[-1]
        self.bucket = self.client.bucket(self.bucket_name)
        yield self.bucket.enable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.bucket.disable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def test_set_data_empty(self):
        """Get an object that does not exist, then set_data and save it """
        log.msg("*** set_data_empty")

        obj = yield self.bucket.get("foo1")

        self.assertEqual(obj.exists(), False)
        self.assertEqual(obj.get_data(), None)

        obj.set_data('bar1')
        yield obj.store()

        self.assertEqual(obj.exists(), True)
        self.assertEqual(obj.get_data(), "bar1")
        log.msg("done set_data_empty")

    @defer.inlineCallbacks
    def test_add_and_delete(self):
        """Basic adds and deletes"""
        log.msg("*** add_and_delete")

        obj = self.bucket.new("foo1", "test1")
        yield obj.store()

        self.assertEqual(obj.exists(), True)
        self.assertEqual(obj.get_data(), "test1")

        obj.set_data('bar1')
        yield obj.store()

        obj = yield self.bucket.get("foo1")
        self.assertEqual(obj.exists(), True)
        self.assertEqual(obj.get_data(), "bar1")

        yield obj.delete()

        obj = yield self.bucket.get("foo1")
        self.assertEqual(obj.exists(), False)
        log.msg("done add_and_delete")

    @defer.inlineCallbacks
    def test_list_keys(self):
        """Test listing all keys in bucket."""
        log.msg("*** list_keys")

        obj = self.bucket.new("foo1", "test1")
        yield obj.store()
        obj1 = self.bucket.new("foo2", "test2")
        yield obj1.store()

        keys = yield self.bucket.list_keys()
        self.assertEqual([u"foo1", u"foo2"], sorted(keys))

    @defer.inlineCallbacks
    def test_purge_keys(self):
        """Test purging all keys in a bucket."""
        log.msg("*** purge_keys")

        obj = self.bucket.new("foo1", "test1")
        yield obj.store()
        obj1 = self.bucket.new("foo2", "test2")
        yield obj1.store()

        yield self.bucket.purge_keys()

        # FIXME: nasty hack to work around purge_keys returning
        # too soon (or maybe list_keys being weird).
        import time
        start = time.time()
        while True:
            keys = yield self.bucket.list_keys()
            if not keys or time.time() - start > 10:
                break
            time.sleep(0.2)
        self.assertEqual([], keys)

    @defer.inlineCallbacks
    def test_store_and_get(self):
        """Store and get text data."""
        log.msg('*** store_and_get')
        data = 'blueprint'
        obj = self.bucket.new('blue_foo1', data)
        yield obj.store()
        del obj

        obj1 = yield self.bucket.get('blue_foo1')
        self.assertEqual(obj1.exists(), True)
        self.assertEqual(obj1.get_bucket().get_name(), self.bucket_name)
        self.assertEqual(obj1.get_key(), 'blue_foo1')
        self.assertEqual(obj1.get_data(), data)
        log.msg('done store_and_get')

    @defer.inlineCallbacks
    def test_binary_store_and_get(self):
        """store and get binary data."""

        log.msg('*** binary_store_and_get')

        # Store as binary, retrieve as binary, then compare...
        rand = str(randint())
        obj = self.bucket.new_binary('foo1', rand)
        yield obj.store()
        del obj

        obj = yield self.bucket.get_binary('foo1')
        self.assertEqual(obj.exists(), True)
        self.assertEqual(obj.get_data(), rand)
        del obj

        # Store as JSON, retrieve as binary, JSON-decode, then compare...
        data = [randint(), randint(), randint()]
        obj = self.bucket.new('foo2', data)
        yield obj.store()
        del obj

        obj = yield self.bucket.get_binary('foo2')
        self.assertEqual(data, json.loads(obj.get_data()))
        log.msg('done binary_store_and_get')

    @defer.inlineCallbacks
    def test_missing_object(self):
        """handle missing objects."""
        log.msg('*** missing_object')
        obj = yield self.bucket.get("missing")
        self.assertEqual(not obj.exists(), True)
        self.assertEqual(obj.get_data(), None)
        log.msg('done missing_object')

    @defer.inlineCallbacks
    def test_delete(self):
        """delete objects"""
        log.msg('*** delete')
        rand = randint()
        obj = self.bucket.new('foo', rand)
        yield obj.store()
        obj = yield self.bucket.get('foo')
        self.assertEqual(obj.exists(), True)
        yield obj.delete()
        yield obj.reload()
        self.assertEqual(obj.exists(), False)
        log.msg('done delete')

    @defer.inlineCallbacks
    def test_siblings(self):
        """find siblings"""
        # Siblings works except for the store at the end.
        # Need to isolate and test that separately.
        log.msg('*** siblings')

        # Set up the bucket, clear any existing object...
        yield self.bucket.set_allow_multiples(True)
        obj = yield self.bucket.get('foo')
        yield obj.delete()

        obj = yield self.bucket.get('foo')
        self.assertEqual(obj.exists(), False)

        # Store the same object multiple times...
        for i in range(5):
            # for this to work, we must get a new RiakClient
            # on each pass and it must have a different client_id.
            # calling RiakClient without params uses randomly-generate id.
            client = riak.RiakClient()
            bucket = client.bucket(self.bucket_name)
            yield bucket.new('foo', randint()).store()

        # Make sure the object has 5 siblings...
        yield obj.reload()
        self.assertEqual(obj.has_siblings(), True)
        self.assertEqual(obj.get_sibling_count(), 5)

        # Test get_sibling()/get_siblings()...
        siblings = yield obj.get_siblings()
        obj3 = yield obj.get_sibling(3)

        self.assertEqual(siblings[3].get_data(), obj3.get_data())

        # Resolve the conflict, and then do a get...
        obj3 = yield obj.get_sibling(3)
        yield obj3.store()
        yield obj.reload()
        self.assertEqual(obj.get_data(), obj3.get_data())

        # Clean up for next test...
        yield obj.delete()
        log.msg('done siblings')

    @defer.inlineCallbacks
    def test_store_and_get_links(self):
        """manipulate links"""
        # if we store with the basho test, we get this correctly.
        # so there's now something wrong with link storage.
        log.msg('*** store_and_get_links')

        quote_string = "tag2!@#%^&*)"

        obj = self.bucket.new("foo", 2) \
                .add_link(self.bucket.new("foo1")) \
                .add_link(self.bucket.new("foo2"), "tag") \
                .add_link(self.bucket.new(quote_string), quote_string)
        yield obj.store()
        del obj

        log.msg("Get the Links")
        obj = yield self.bucket.get("foo")
        links = obj.get_links()
        self.assertEqual(len(links), 3)

        quote_link = links[-1]
        self.assertEqual(quote_link.get_key(), quote_string)
        self.assertEqual(quote_link.get_tag(), quote_string)

        log.msg('done store_and_get_links')

    @defer.inlineCallbacks
    def test_link_walking(self):
        """walk links"""
        log.msg('*** link_walking')
        obj_1 = self.bucket.new("foo1", "test1")
        yield obj_1.store()

        obj_2 = self.bucket.new("foo2", "test2")
        yield obj_2.store()

        obj_3 = self.bucket.new("foo3", "test3")
        yield obj_3.store()

        obj = self.bucket.new("foo", 2) \
                .add_link(obj_1) \
                .add_link(obj_2, "tag") \
                .add_link(obj_3, "tag2!@#%^&*)")
        yield obj.store()
        obj = yield self.bucket.get("foo")
        job = obj.link(self.bucket_name)
        results = yield job.run()
        self.assertEqual(len(results), 3)
        results = yield obj.link(self.bucket_name, "tag").run()
        self.assertEqual(len(results), 1)
        log.msg('done link_walking')

    @defer.inlineCallbacks
    def test_meta_data_simple(self):
        """ensure we can get and set metadata"""
        log.msg('*** meta_data_simple')

        key = "foo1"
        key_data = "test1"

        # be sure object is deleted before we start
        obj = yield self.bucket.get(key)
        yield obj.delete()

        # now get a fresh new one
        obj = self.bucket.new(key, key_data)

        # see we can store and get back a header
        meta_key = 'this-is-a-test'
        meta_key1 = 'Fifty-Three'
        meta_data = 'ABCDEFG 123'
        obj.add_meta_data(meta_key, meta_data)
        metas = obj.get_all_meta_data()
        self.assertTrue(meta_key in metas)
        self.assertEqual(1, len(metas))

        # add second meta_data and get back both.
        # since they are stored as dictionary, we cannot
        # trus the order
        obj.add_meta_data(meta_key1, meta_data)
        metas = obj.get_all_meta_data()
        self.assertEqual(2, len(metas))
        self.assertTrue(meta_key in metas)
        self.assertTrue(meta_key1 in metas)

        # now delete one of the metas
        obj.remove_meta_data(meta_key)
        metas = obj.get_all_meta_data()
        self.assertFalse(meta_key in metas)
        self.assertTrue(meta_key1 in metas)
        self.assertEqual(1, len(metas))

        # store the object
        # since return body is true, the obj is refreshed
        yield obj.store()
        metas = obj.get_all_meta_data()

        # remember: twisted lowercases the headers!
        mk1_lc = meta_key1.lower()
        self.assertTrue(mk1_lc in metas)
        self.assertTrue(metas[mk1_lc] == meta_data)
        log.msg('done meta_data_simple')



# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)


DESCRIPTOR = descriptor.FileDescriptor(
  name='riak_kv.proto',
  package='',
  serialized_pb='\n\rriak_kv.proto\x1a\nriak.proto\"\'\n\x12RpbGetClientIdResp\x12\x11\n\tclient_id\x18\x01 \x02(\x0c\"&\n\x11RpbSetClientIdReq\x12\x11\n\tclient_id\x18\x01 \x02(\x0c\"\xa4\x01\n\tRpbGetReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\x12\x0b\n\x03key\x18\x02 \x02(\x0c\x12\t\n\x01r\x18\x03 \x01(\r\x12\n\n\x02pr\x18\x04 \x01(\r\x12\x14\n\x0c\x62\x61sic_quorum\x18\x05 \x01(\x08\x12\x13\n\x0bnotfound_ok\x18\x06 \x01(\x08\x12\x13\n\x0bif_modified\x18\x07 \x01(\x0c\x12\x0c\n\x04head\x18\x08 \x01(\x08\x12\x15\n\rdeletedvclock\x18\t \x01(\x08\"M\n\nRpbGetResp\x12\x1c\n\x07\x63ontent\x18\x01 \x03(\x0b\x32\x0b.RpbContent\x12\x0e\n\x06vclock\x18\x02 \x01(\x0c\x12\x11\n\tunchanged\x18\x03 \x01(\x08\"\xd3\x01\n\tRpbPutReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\x12\x0b\n\x03key\x18\x02 \x01(\x0c\x12\x0e\n\x06vclock\x18\x03 \x01(\x0c\x12\x1c\n\x07\x63ontent\x18\x04 \x02(\x0b\x32\x0b.RpbContent\x12\t\n\x01w\x18\x05 \x01(\r\x12\n\n\x02\x64w\x18\x06 \x01(\r\x12\x13\n\x0breturn_body\x18\x07 \x01(\x08\x12\n\n\x02pw\x18\x08 \x01(\r\x12\x17\n\x0fif_not_modified\x18\t \x01(\x08\x12\x15\n\rif_none_match\x18\n \x01(\x08\x12\x13\n\x0breturn_head\x18\x0b \x01(\x08\"G\n\nRpbPutResp\x12\x1c\n\x07\x63ontent\x18\x01 \x03(\x0b\x32\x0b.RpbContent\x12\x0e\n\x06vclock\x18\x02 \x01(\x0c\x12\x0b\n\x03key\x18\x03 \x01(\x0c\"~\n\tRpbDelReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\x12\x0b\n\x03key\x18\x02 \x02(\x0c\x12\n\n\x02rw\x18\x03 \x01(\r\x12\x0e\n\x06vclock\x18\x04 \x01(\x0c\x12\t\n\x01r\x18\x05 \x01(\r\x12\t\n\x01w\x18\x06 \x01(\r\x12\n\n\x02pr\x18\x07 \x01(\r\x12\n\n\x02pw\x18\x08 \x01(\r\x12\n\n\x02\x64w\x18\t \x01(\r\"%\n\x12RpbListBucketsResp\x12\x0f\n\x07\x62uckets\x18\x01 \x03(\x0c\" \n\x0eRpbListKeysReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\"-\n\x0fRpbListKeysResp\x12\x0c\n\x04keys\x18\x01 \x03(\x0c\x12\x0c\n\x04\x64one\x18\x02 \x01(\x08\"!\n\x0fRpbGetBucketReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\"2\n\x10RpbGetBucketResp\x12\x1e\n\x05props\x18\x01 \x02(\x0b\x32\x0f.RpbBucketProps\"A\n\x0fRpbSetBucketReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\x12\x1e\n\x05props\x18\x02 \x02(\x0b\x32\x0f.RpbBucketProps\"5\n\x0cRpbMapRedReq\x12\x0f\n\x07request\x18\x01 \x02(\x0c\x12\x14\n\x0c\x63ontent_type\x18\x02 \x02(\x0c\">\n\rRpbMapRedResp\x12\r\n\x05phase\x18\x01 \x01(\r\x12\x10\n\x08response\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64one\x18\x03 \x01(\x08\"\xb0\x01\n\x0bRpbIndexReq\x12\x0e\n\x06\x62ucket\x18\x01 \x02(\x0c\x12\r\n\x05index\x18\x02 \x02(\x0c\x12*\n\x05qtype\x18\x03 \x02(\x0e\x32\x1b.RpbIndexReq.IndexQueryType\x12\x0b\n\x03key\x18\x04 \x01(\x0c\x12\x11\n\trange_min\x18\x05 \x01(\x0c\x12\x11\n\trange_max\x18\x06 \x01(\x0c\"#\n\x0eIndexQueryType\x12\x06\n\x02\x65q\x10\x00\x12\t\n\x05range\x10\x01\"\x1c\n\x0cRpbIndexResp\x12\x0c\n\x04keys\x18\x01 \x03(\x0c\"\xf5\x01\n\nRpbContent\x12\r\n\x05value\x18\x01 \x02(\x0c\x12\x14\n\x0c\x63ontent_type\x18\x02 \x01(\x0c\x12\x0f\n\x07\x63harset\x18\x03 \x01(\x0c\x12\x18\n\x10\x63ontent_encoding\x18\x04 \x01(\x0c\x12\x0c\n\x04vtag\x18\x05 \x01(\x0c\x12\x17\n\x05links\x18\x06 \x03(\x0b\x32\x08.RpbLink\x12\x10\n\x08last_mod\x18\x07 \x01(\r\x12\x16\n\x0elast_mod_usecs\x18\x08 \x01(\r\x12\x1a\n\x08usermeta\x18\t \x03(\x0b\x32\x08.RpbPair\x12\x19\n\x07indexes\x18\n \x03(\x0b\x32\x08.RpbPair\x12\x0f\n\x07\x64\x65leted\x18\x0b \x01(\x08\"3\n\x07RpbLink\x12\x0e\n\x06\x62ucket\x18\x01 \x01(\x0c\x12\x0b\n\x03key\x18\x02 \x01(\x0c\x12\x0b\n\x03tag\x18\x03 \x01(\x0c\"3\n\x0eRpbBucketProps\x12\r\n\x05n_val\x18\x01 \x01(\r\x12\x12\n\nallow_mult\x18\x02 \x01(\x08\x42#\n\x17\x63om.basho.riak.protobufB\x08RiakKvPB')



_RPBINDEXREQ_INDEXQUERYTYPE = descriptor.EnumDescriptor(
  name='IndexQueryType',
  full_name='RpbIndexReq.IndexQueryType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    descriptor.EnumValueDescriptor(
      name='eq', index=0, number=0,
      options=None,
      type=None),
    descriptor.EnumValueDescriptor(
      name='range', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1306,
  serialized_end=1341,
)


_RPBGETCLIENTIDRESP = descriptor.Descriptor(
  name='RpbGetClientIdResp',
  full_name='RpbGetClientIdResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='client_id', full_name='RpbGetClientIdResp.client_id', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=29,
  serialized_end=68,
)


_RPBSETCLIENTIDREQ = descriptor.Descriptor(
  name='RpbSetClientIdReq',
  full_name='RpbSetClientIdReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='client_id', full_name='RpbSetClientIdReq.client_id', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=70,
  serialized_end=108,
)


_RPBGETREQ = descriptor.Descriptor(
  name='RpbGetReq',
  full_name='RpbGetReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='bucket', full_name='RpbGetReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='key', full_name='RpbGetReq.key', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='r', full_name='RpbGetReq.r', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='pr', full_name='RpbGetReq.pr', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='basic_quorum', full_name='RpbGetReq.basic_quorum', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='notfound_ok', full_name='RpbGetReq.notfound_ok', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='if_modified', full_name='RpbGetReq.if_modified', index=6,
      number=7, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='head', full_name='RpbGetReq.head', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='deletedvclock', full_name='RpbGetReq.deletedvclock', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=111,
  serialized_end=275,
)


_RPBGETRESP = descriptor.Descriptor(
  name='RpbGetResp',
  full_name='RpbGetResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='content', full_name='RpbGetResp.content', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='vclock', full_name='RpbGetResp.vclock', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='unchanged', full_name='RpbGetResp.unchanged', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=277,
  serialized_end=354,
)


_RPBPUTREQ = descriptor.Descriptor(
  name='RpbPutReq',
  full_name='RpbPutReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='bucket', full_name='RpbPutReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='key', full_name='RpbPutReq.key', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='vclock', full_name='RpbPutReq.vclock', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='content', full_name='RpbPutReq.content', index=3,
      number=4, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='w', full_name='RpbPutReq.w', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='dw', full_name='RpbPutReq.dw', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='return_body', full_name='RpbPutReq.return_body', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='pw', full_name='RpbPutReq.pw', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='if_not_modified', full_name='RpbPutReq.if_not_modified', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='if_none_match', full_name='RpbPutReq.if_none_match', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='return_head', full_name='RpbPutReq.return_head', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=357,
  serialized_end=568,
)


_RPBPUTRESP = descriptor.Descriptor(
  name='RpbPutResp',
  full_name='RpbPutResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='content', full_name='RpbPutResp.content', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='vclock', full_name='RpbPutResp.vclock', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='key', full_name='RpbPutResp.key', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=570,
  serialized_end=641,
)


_RPBDELREQ = descriptor.Descriptor(
  name='RpbDelReq',
  full_name='RpbDelReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='bucket', full_name='RpbDelReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='key', full_name='RpbDelReq.key', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='rw', full_name='RpbDelReq.rw', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='vclock', full_name='RpbDelReq.vclock', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='r', full_name='RpbDelReq.r', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='w', full_name='RpbDelReq.w', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='pr', full_name='RpbDelReq.pr', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='pw', full_name='RpbDelReq.pw', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='dw', full_name='RpbDelReq.dw', index=8,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=643,
  serialized_end=769,
)


_RPBLISTBUCKETSRESP = descriptor.Descriptor(
  name='RpbListBucketsResp',
  full_name='RpbListBucketsResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='buckets', full_name='RpbListBucketsResp.buckets', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=771,
  serialized_end=808,
)


_RPBLISTKEYSREQ = descriptor.Descriptor(
  name='RpbListKeysReq',
  full_name='RpbListKeysReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='bucket', full_name='RpbListKeysReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=810,
  serialized_end=842,
)


_RPBLISTKEYSRESP = descriptor.Descriptor(
  name='RpbListKeysResp',
  full_name='RpbListKeysResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='keys', full_name='RpbListKeysResp.keys', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='done', full_name='RpbListKeysResp.done', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=844,
  serialized_end=889,
)


_RPBGETBUCKETREQ = descriptor.Descriptor(
  name='RpbGetBucketReq',
  full_name='RpbGetBucketReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='bucket', full_name='RpbGetBucketReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=891,
  serialized_end=924,
)


_RPBGETBUCKETRESP = descriptor.Descriptor(
  name='RpbGetBucketResp',
  full_name='RpbGetBucketResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='props', full_name='RpbGetBucketResp.props', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=926,
  serialized_end=976,
)


_RPBSETBUCKETREQ = descriptor.Descriptor(
  name='RpbSetBucketReq',
  full_name='RpbSetBucketReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='bucket', full_name='RpbSetBucketReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='props', full_name='RpbSetBucketReq.props', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=978,
  serialized_end=1043,
)


_RPBMAPREDREQ = descriptor.Descriptor(
  name='RpbMapRedReq',
  full_name='RpbMapRedReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='request', full_name='RpbMapRedReq.request', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='content_type', full_name='RpbMapRedReq.content_type', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1045,
  serialized_end=1098,
)


_RPBMAPREDRESP = descriptor.Descriptor(
  name='RpbMapRedResp',
  full_name='RpbMapRedResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='phase', full_name='RpbMapRedResp.phase', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='response', full_name='RpbMapRedResp.response', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='done', full_name='RpbMapRedResp.done', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1100,
  serialized_end=1162,
)


_RPBINDEXREQ = descriptor.Descriptor(
  name='RpbIndexReq',
  full_name='RpbIndexReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='bucket', full_name='RpbIndexReq.bucket', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='index', full_name='RpbIndexReq.index', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='qtype', full_name='RpbIndexReq.qtype', index=2,
      number=3, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='key', full_name='RpbIndexReq.key', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='range_min', full_name='RpbIndexReq.range_min', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='range_max', full_name='RpbIndexReq.range_max', index=5,
      number=6, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RPBINDEXREQ_INDEXQUERYTYPE,
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1165,
  serialized_end=1341,
)


_RPBINDEXRESP = descriptor.Descriptor(
  name='RpbIndexResp',
  full_name='RpbIndexResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='keys', full_name='RpbIndexResp.keys', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1343,
  serialized_end=1371,
)


_RPBCONTENT = descriptor.Descriptor(
  name='RpbContent',
  full_name='RpbContent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='value', full_name='RpbContent.value', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='content_type', full_name='RpbContent.content_type', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='charset', full_name='RpbContent.charset', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='content_encoding', full_name='RpbContent.content_encoding', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='vtag', full_name='RpbContent.vtag', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='links', full_name='RpbContent.links', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='last_mod', full_name='RpbContent.last_mod', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='last_mod_usecs', full_name='RpbContent.last_mod_usecs', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='usermeta', full_name='RpbContent.usermeta', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='indexes', full_name='RpbContent.indexes', index=9,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='deleted', full_name='RpbContent.deleted', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1374,
  serialized_end=1619,
)


_RPBLINK = descriptor.Descriptor(
  name='RpbLink',
  full_name='RpbLink',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='bucket', full_name='RpbLink.bucket', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='key', full_name='RpbLink.key', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='tag', full_name='RpbLink.tag', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1621,
  serialized_end=1672,
)


_RPBBUCKETPROPS = descriptor.Descriptor(
  name='RpbBucketProps',
  full_name='RpbBucketProps',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='n_val', full_name='RpbBucketProps.n_val', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='allow_mult', full_name='RpbBucketProps.allow_mult', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1674,
  serialized_end=1725,
)

import riak_pb2

_RPBGETRESP.fields_by_name['content'].message_type = _RPBCONTENT
_RPBPUTREQ.fields_by_name['content'].message_type = _RPBCONTENT
_RPBPUTRESP.fields_by_name['content'].message_type = _RPBCONTENT
_RPBGETBUCKETRESP.fields_by_name['props'].message_type = _RPBBUCKETPROPS
_RPBSETBUCKETREQ.fields_by_name['props'].message_type = _RPBBUCKETPROPS
_RPBINDEXREQ.fields_by_name['qtype'].enum_type = _RPBINDEXREQ_INDEXQUERYTYPE
_RPBINDEXREQ_INDEXQUERYTYPE.containing_type = _RPBINDEXREQ;
_RPBCONTENT.fields_by_name['links'].message_type = _RPBLINK
_RPBCONTENT.fields_by_name['usermeta'].message_type = riak_pb2._RPBPAIR
_RPBCONTENT.fields_by_name['indexes'].message_type = riak_pb2._RPBPAIR

class RpbGetClientIdResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBGETCLIENTIDRESP
  
  # @@protoc_insertion_point(class_scope:RpbGetClientIdResp)

class RpbSetClientIdReq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBSETCLIENTIDREQ
  
  # @@protoc_insertion_point(class_scope:RpbSetClientIdReq)

class RpbGetReq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBGETREQ
  
  # @@protoc_insertion_point(class_scope:RpbGetReq)

class RpbGetResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBGETRESP
  
  # @@protoc_insertion_point(class_scope:RpbGetResp)

class RpbPutReq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBPUTREQ
  
  # @@protoc_insertion_point(class_scope:RpbPutReq)

class RpbPutResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBPUTRESP
  
  # @@protoc_insertion_point(class_scope:RpbPutResp)

class RpbDelReq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBDELREQ
  
  # @@protoc_insertion_point(class_scope:RpbDelReq)

class RpbListBucketsResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBLISTBUCKETSRESP
  
  # @@protoc_insertion_point(class_scope:RpbListBucketsResp)

class RpbListKeysReq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBLISTKEYSREQ
  
  # @@protoc_insertion_point(class_scope:RpbListKeysReq)

class RpbListKeysResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBLISTKEYSRESP
  
  # @@protoc_insertion_point(class_scope:RpbListKeysResp)

class RpbGetBucketReq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBGETBUCKETREQ
  
  # @@protoc_insertion_point(class_scope:RpbGetBucketReq)

class RpbGetBucketResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBGETBUCKETRESP
  
  # @@protoc_insertion_point(class_scope:RpbGetBucketResp)

class RpbSetBucketReq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBSETBUCKETREQ
  
  # @@protoc_insertion_point(class_scope:RpbSetBucketReq)

class RpbMapRedReq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBMAPREDREQ
  
  # @@protoc_insertion_point(class_scope:RpbMapRedReq)

class RpbMapRedResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBMAPREDRESP
  
  # @@protoc_insertion_point(class_scope:RpbMapRedResp)

class RpbIndexReq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBINDEXREQ
  
  # @@protoc_insertion_point(class_scope:RpbIndexReq)

class RpbIndexResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBINDEXRESP
  
  # @@protoc_insertion_point(class_scope:RpbIndexResp)

class RpbContent(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBCONTENT
  
  # @@protoc_insertion_point(class_scope:RpbContent)

class RpbLink(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBLINK
  
  # @@protoc_insertion_point(class_scope:RpbLink)

class RpbBucketProps(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBBUCKETPROPS
  
  # @@protoc_insertion_point(class_scope:RpbBucketProps)

# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
"""
riakasaurus trial test file.
riakasaurus _must_ be on your PYTHONPATH

"""

import json
import random
from twisted.trial import unittest
from twisted.python import log
from twisted.internet import defer, reactor

VERBOSE = False

from riakasaurus import riak

# uncomment to activate logging
# import sys
# log.startLogging(sys.stderr)

RIAK_CLIENT_ID = 'TEST'
BUCKET_PREFIX = 'riakasaurus.tests.'

JAVASCRIPT_SUM = """
function(v) {
  x = v.reduce(function(a,b){ return a + b }, 0);
  return [x];
}
"""


def sleep(secs):
    d = defer.Deferred()
    reactor.callLater(secs, d.callback, None)
    return d


def randint():
    """Generate nice random int for our test."""
    return random.randint(1, 999999)


class Tests(unittest.TestCase):
    """
    trial unit tests.
    """

    test_keys = ['foo', 'foo1', 'foo2', 'foo3', 'bar', 'baz', 'ba_foo1',
                 'blue_foo1']

    @defer.inlineCallbacks
    def setUp(self):
        self.client = riak.RiakClient(client_id=RIAK_CLIENT_ID)
        self.bucket_name = BUCKET_PREFIX + self.id().rsplit('.', 1)[-1]
        self.bucket = self.client.bucket(self.bucket_name)
        yield self.bucket.enable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.bucket.disable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def test_riak_search(self):
        """Test searching buckets"""
        log.msg("*** riak_search")

        yield self.bucket.enable_search()

        se = yield self.bucket.search_enabled()
        self.assertEqual(se, True)

        obj1 = self.bucket.new("foo1", {"foo": "test1"})
        yield obj1.store()

        s = self.client.search(self.bucket_name, 'foo:test1')
        keys = yield s.run()

        v1 = yield keys[0].get()

        self.assertTrue(v1.get_key() == u'foo1')

        yield obj1.delete()
        yield self.bucket.disable_search()

    @defer.inlineCallbacks
    def test_search_enable_disable(self):
        yield self.bucket.disable_search()
        self.assertEqual((yield self.bucket.search_enabled()), False)
        yield self.bucket.enable_search()
        self.assertEqual((yield self.bucket.search_enabled()), True)
        yield self.bucket.disable_search()
        self.assertEqual((yield self.bucket.search_enabled()), False)

    @defer.inlineCallbacks
    def test_solr_search_from_bucket(self):
        yield self.bucket.new("user", {"username": "roidrage"}).store()
        results = yield self.bucket.search("username:roidrage")
        self.assertEquals(1, len(results["docs"]))

    @defer.inlineCallbacks
    def test_solr_search_with_params_from_bucket(self):
        yield self.bucket.new("user", {"username": "roidrage"}).store()
        result = yield self.bucket.search("username:roidrage", wt="xml")

        self.assertEquals(1, len(result["docs"]))

    @defer.inlineCallbacks
    def test_solr_search_with_params(self):
        yield self.bucket.new("user", {"username": "roidrage"}).store()
        result = yield self.client.solr().search(self.bucket_name,
                                                  "username:roidrage",
                                                  wt="xml")

        self.assertEquals(1, len(result["docs"]))

    @defer.inlineCallbacks
    def test_solr_search(self):
        yield self.bucket.new("user", {"username": "roidrage"}).store()
        results = yield self.client.solr().search(self.bucket_name,
                                                  "username:roidrage")
        self.assertEquals(1, len(results["docs"]))

    @defer.inlineCallbacks
    def test_add_document_to_index(self):
        yield self.client.solr().add(self.bucket_name,
                                     {"id": "doc", "username": "tony"})
        results = yield self.client.solr().search(self.bucket_name,
                                                  "username:tony")

        self.assertEquals("tony",
                          results["docs"][0]["username"])

    @defer.inlineCallbacks
    def test_add_multiple_documents_to_index(self):
        yield self.client.solr().add(self.bucket_name,
                                     {"id": "dizzy", "username": "dizzy"},
                                     {"id": "russell", "username": "russell"})
        yield sleep(2)  # Eventual consistency is annoying
        results = yield self.client.solr().search(self.bucket_name,
                                                  "username:russell OR"
                                                  " username:dizzy")
        self.assertEquals(2, len(results["docs"]))

    @defer.inlineCallbacks
    def test_delete_documents_from_search_by_id(self):
        yield self.client.solr().add(self.bucket_name,
                                     {"id": "dizzy", "username": "dizzy"},
                                     {"id": "russell", "username": "russell"})
        yield self.client.solr().delete(self.bucket_name, docs=["dizzy"])
        results = yield self.client.solr().search(self.bucket_name,
                                                  "username:russell OR"
                                                  " username:dizzy")
        # This test fails at eventual consistency...
        #self.assertEquals(1, len(results["docs"]))

    @defer.inlineCallbacks
    def test_delete_documents_from_search_by_query(self):
        yield self.client.solr().add(self.bucket_name,
                                     {"id": "dizzy", "username": "dizzy"},
                                     {"id": "russell", "username": "russell"})
        yield self.client.solr().delete(self.bucket_name,
                                        queries=["username:dizzy",
                                                 "username:russell"])
        results = yield self.client.solr().search(self.bucket_name,
                                                  "username:russell OR"
                                                  " username:dizzy")
        # This test fails at eventual consistency...
        #self.assertEquals(0, len(results["docs"]))



from twisted.internet import defer, reactor
from twisted.python import log

from test_objects import Tests
from riakasaurus import riak, transport

RIAK_CLIENT_ID = 'TEST'
BUCKET_PREFIX = 'riakasaurus.tests.'

VERBOSE = False
# uncomment to activate logging
if VERBOSE:
    import sys
    log.startLogging(sys.stderr)


class Tests_PB(Tests):

    @defer.inlineCallbacks
    def setUp(self):
        self.client = riak.RiakClient(client_id=RIAK_CLIENT_ID,
                                      host='127.0.0.1',
                                      port=8087,
                                      transport=transport.PBCTransport)
        # self.client.debug = 0
        # self.client.get_transport().debug = 0
        self.bucket_name = BUCKET_PREFIX + self.id().rsplit('.', 1)[-1]
        self.bucket = self.client.bucket(self.bucket_name)
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def tearDown(self):
        # shut down pb connection explicitly
        yield self.client.get_transport().quit()

    @defer.inlineCallbacks
    def test_meta_data_simple(self):
        """
        differs from http twisted in that key case is
        maintained by PB, but not via HTTP
        """
        log.msg('*** meta_data_simple')

        key = "foo1"
        key_data = "test1"

        # be sure object is deleted before we start
        obj = yield self.bucket.get(key)
        yield obj.delete()

        # now get a fresh new one
        obj = self.bucket.new(key, key_data)

        # see we can store and get back a header
        meta_key = 'this-is-a-test'
        meta_key1 = 'Fifty-Three'
        meta_data = 'ABCDEFG 123'
        obj.add_meta_data(meta_key, meta_data)
        metas = obj.get_all_meta_data()
        self.assertTrue(meta_key in metas)
        self.assertEqual(1, len(metas))

        # add second meta_data and get back both.
        # since they are stored as dictionary, we cannot
        # trus the order
        obj.add_meta_data(meta_key1, meta_data)
        metas = obj.get_all_meta_data()
        self.assertEqual(2, len(metas))
        self.assertTrue(meta_key in metas)
        self.assertTrue(meta_key1 in metas)

        # now delete one of the metas
        obj.remove_meta_data(meta_key)
        metas = obj.get_all_meta_data()
        self.assertFalse(meta_key in metas)
        self.assertTrue(meta_key1 in metas)
        self.assertEqual(1, len(metas))

        # store the object
        # since return body is true, the obj is refreshed
        yield obj.store()
        metas = obj.get_all_meta_data()

        # differs from http, PB maintains case
        self.assertTrue(meta_key1 in metas)
        self.assertTrue(metas[meta_key1] == meta_data)
        log.msg('done meta_data_simple')

    def test_link_walking(self):
        """mapred not supported yet, so skip test"""
        self.assertTrue(True)



from twisted.internet import defer, reactor
from twisted.python import log
from twisted.trial import unittest

from test_basic import BasicTestsMixin
from riakasaurus import riak, transport

RIAK_CLIENT_ID = 'TEST'
BUCKET_PREFIX = 'riakasaurus.tests.'

VERBOSE = False
# uncomment to activate logging
if VERBOSE:
    import sys
    log.startLogging(sys.stderr)


class Tests_PB(unittest.TestCase, BasicTestsMixin):

    @defer.inlineCallbacks
    def setUp(self):
        self.client = riak.RiakClient(client_id=RIAK_CLIENT_ID,
                                      host='127.0.0.1',
                                      port=8087,
                                      transport=transport.PBCTransport)
        # self.client.debug = 0
        # self.client.get_transport().debug = 0
        self.bucket_name = BUCKET_PREFIX + self.id().rsplit('.', 1)[-1]
        self.bucket = self.client.bucket(self.bucket_name)
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def tearDown(self):
        # shut down pb connection explicitly
        yield self.client.get_transport().quit()

    @defer.inlineCallbacks
    def test_is_alive(self):
        alive = yield self.client.is_alive()
        self.assertEqual(alive, True)



# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)


DESCRIPTOR = descriptor.FileDescriptor(
  name='riak.proto',
  package='',
  serialized_pb='\n\nriak.proto\"/\n\x0cRpbErrorResp\x12\x0e\n\x06\x65rrmsg\x18\x01 \x02(\x0c\x12\x0f\n\x07\x65rrcode\x18\x02 \x02(\r\"<\n\x14RpbGetServerInfoResp\x12\x0c\n\x04node\x18\x01 \x01(\x0c\x12\x16\n\x0eserver_version\x18\x02 \x01(\x0c\"%\n\x07RpbPair\x12\x0b\n\x03key\x18\x01 \x02(\x0c\x12\r\n\x05value\x18\x02 \x01(\x0c\x42!\n\x17\x63om.basho.riak.protobufB\x06RiakPB')




_RPBERRORRESP = descriptor.Descriptor(
  name='RpbErrorResp',
  full_name='RpbErrorResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='errmsg', full_name='RpbErrorResp.errmsg', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='errcode', full_name='RpbErrorResp.errcode', index=1,
      number=2, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=14,
  serialized_end=61,
)


_RPBGETSERVERINFORESP = descriptor.Descriptor(
  name='RpbGetServerInfoResp',
  full_name='RpbGetServerInfoResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='node', full_name='RpbGetServerInfoResp.node', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='server_version', full_name='RpbGetServerInfoResp.server_version', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=63,
  serialized_end=123,
)


_RPBPAIR = descriptor.Descriptor(
  name='RpbPair',
  full_name='RpbPair',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='key', full_name='RpbPair.key', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='value', full_name='RpbPair.value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=125,
  serialized_end=162,
)



class RpbErrorResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBERRORRESP
  
  # @@protoc_insertion_point(class_scope:RpbErrorResp)

class RpbGetServerInfoResp(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBGETSERVERINFORESP
  
  # @@protoc_insertion_point(class_scope:RpbGetServerInfoResp)

class RpbPair(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _RPBPAIR
  
  # @@protoc_insertion_point(class_scope:RpbPair)

# @@protoc_insertion_point(module_scope)



"""Mostly a compatibility module"""

from riakasaurus import riak_object, mapreduce, bucket, client

RiakClient = client.RiakClient
RiakBucket = bucket.RiakBucket
RiakObject = riak_object.RiakObject
RiakMapReduce = mapreduce.RiakMapReduce
RiakLink = mapreduce.RiakLink



#!/usr/bin/env python

from twisted.trial import unittest
from twisted.python import log
from twisted.internet import defer
from distutils.version import LooseVersion

from riakasaurus import riak, exceptions

RIAK_CLIENT_ID = 'TEST'
BUCKET_PREFIX = 'riakasaurus.tests.'

class RiakTest(unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.client = riak.RiakClient(client_id=RIAK_CLIENT_ID)
        self.bucket_name = BUCKET_PREFIX + self.id().rsplit('.', 1)[-1]
        self.bucket = self.client.bucket(self.bucket_name)
        yield self.bucket.enable_search()
        yield self.bucket.purge_keys()

def buildTestCases(name):
    tests = [
        'self.client.is_alive',
        'self.client.ping',
        'self.client.stream_buckets',
        'self.client.bucket',
        'self.client.get_keys',
        'self.client.get',
        'self.client.stream_keys',
        'self.client.put', 
        'self.client.multiget', 
        'self.client.delete', 
        'self.client.mapred', 
        'self.client.solr.add',
        'self.client.solr.delete', 
        'self.client.solr.search',
        'self.client.delete_search_index', 
        'self.client.create_search_schema', 
        'self.client.get_search_schema', 
        'self.client.create_search_index', 
        'self.client.get_index',
        'self.client.index', 
    ]

    testcases = {}
    for test in tests:
        def checkMethod(self, fn=test):
            """ Runs a test lambda operation """
            f = eval(fn)

        n = test.strip('self.').replace('.', '_')
        testcases['test_%s' % n] = checkMethod

    return type(name, (RiakTest,), testcases)

Tests = buildTestCases('TestStuff')



#!/usr/bin/env python
"""
riakasaurus trial test file.
riakasaurus _must_ be on your PYTHONPATH

"""

from twisted.trial import unittest
from twisted.python import log
from twisted.internet import defer
from distutils.version import LooseVersion

VERBOSE = False

from riakasaurus import riak, exceptions

# uncomment to activate logging
# import sys
# log.startLogging(sys.stderr)

RIAK_CLIENT_ID = 'TEST'
BUCKET_PREFIX = 'riakasaurus.tests.'


class BasicTestsMixin(object):

    test_keys = ['foo', 'foo1', 'foo2', 'foo3', 'bar', 'baz', 'ba_foo1',
                 'blue_foo1']

    @defer.inlineCallbacks
    def test_head(self):
        """
            create a object, and retrieve metadata via head(), no content
            is loaded
        """
        log.msg("*** head")

        obj = self.bucket.new("foo1", "test1")
        yield obj.store()

        obj = yield self.bucket.head("foo1")
        self.assertEqual(obj.exists(), True)
        self.assertEqual(obj.get_data(), None)

        log.msg("done head")

    @defer.inlineCallbacks
    def test_list_buckets(self):
        """Test listing all buckets."""
        log.msg("*** list_buckets")

        obj1 = self.bucket.new("foo1", "test1")
        yield obj1.store()

        bucket2_name = "%s2" % self.bucket_name
        bucket2 = self.client.bucket(bucket2_name)
        obj2 = bucket2.new("foo2", "test2")
        yield obj2.store()

        buckets = yield self.client.list_buckets()
        # just check that these two buckets exist in case
        # there are buckets not related to these tests
        self.assertTrue(self.bucket_name in buckets)
        self.assertTrue(bucket2_name in buckets)

        # Cleanup after ourselves
        yield obj1.delete()
        yield obj2.delete()

    @defer.inlineCallbacks
    def test_is_alive(self):
        """Can we ping the riak server."""
        log.msg('*** is_alive')
        client_id = self.client.get_client_id()
        self.assertEqual(client_id, RIAK_CLIENT_ID)
        alive = yield self.client.is_alive()
        self.assertEqual(alive, True)
        log.msg('done is_alive')

    @defer.inlineCallbacks
    def test_set_bucket_properties(self):
        """manipulate bucket properties"""
        log.msg('*** set_bucket_properties')
        # Test setting allow mult...
        yield self.bucket.set_allow_multiples(True)
        is_multiples = yield self.bucket.get_allow_multiples()
        self.assertEqual(is_multiples, True)
        # Test setting nval...
        yield self.bucket.set_n_val(3)
        n_val = yield self.bucket.get_n_val()
        self.assertEqual(n_val, 3)
        # Test setting multiple properties...
        yield self.bucket.set_properties({"allow_mult": False, "n_val": 2})
        is_multiples = yield self.bucket.get_allow_multiples()
        n_val = yield self.bucket.get_n_val()
        self.assertEqual(is_multiples, False)
        self.assertEqual(n_val, 2)
        log.msg('done set_bucket_properties')


class Tests_HTTP(unittest.TestCase, BasicTestsMixin):
    """
    trial unit tests.
    """

    @defer.inlineCallbacks
    def setUp(self):
        self.client = riak.RiakClient(client_id=RIAK_CLIENT_ID)
        self.bucket_name = BUCKET_PREFIX + self.id().rsplit('.', 1)[-1]
        self.bucket = self.client.bucket(self.bucket_name)
        yield self.bucket.enable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.bucket.disable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def test_reset_bucket_properties(self):
        """manipulate bucket properties"""
        log.msg('*** reset_bucket_properties')
        try:
            # Reset bucket properties...
            yield self.bucket.reset_properties()
        except Exception, e:
            if 'not supported' in str(e):
                log.msg('skip reset_bucket_properties')
                raise unittest.SkipTest(
                    'bucket.reset_properties() not supported')
            raise
        # Get default n_val...
        default_n_val = yield self.bucket.get_n_val()
        # Set new n_val...
        yield self.bucket.set_n_val(default_n_val + 1)
        n_val = yield self.bucket.get_n_val()
        self.assertEqual(n_val, default_n_val + 1)
        # Test resetting properties again...
        yield self.bucket.reset_properties()
        n_val = yield self.bucket.get_n_val()
        self.assertEqual(n_val, default_n_val)
        log.msg('done reset_bucket_properties')

    @defer.inlineCallbacks
    def test_reset_bucket_properties_not_available(self):
        """manipulate bucket properties"""
        log.msg('*** reset_bucket_properties_not_available')
        self.patch(self.client.transport, 'server_version',
                   lambda: LooseVersion('1.2.0'))
        try:
            # Test resetting bucket properties...
            yield self.bucket.reset_properties()
            self.fail('Expected "not supported" exception, got nothing.')
        except Exception, e:
            if 'not supported' not in str(e):
                raise
        log.msg('done reset_bucket_properties_not_available')

    @defer.inlineCallbacks
    def test_request_timeout(self):
        """Timeouts are respected."""
        log.msg('*** timeout')

        # Set a very short timeout and expect it to trigger.
        self.client.setRequestTimeout(0.00001)
        try:
            yield self.bucket.get_keys()
            self.fail('Request did not time out.')
        except exceptions.RequestTimeout:
            pass

        # Set a long timeout and expect it not to trigger.
        self.client.setRequestTimeout(1)
        try:
            yield self.bucket.get('foo')
        except exceptions.RequestTimeout:
            self.fail('Request timed out unexpectedly.')

        self.client.setRequestTimeout(None)
        log.msg('done timeout')



"""
Copyright 2010 Rusty Klophaus <rusty@basho.com>
Copyright 2010 Justin Sheehy <justin@basho.com>
Copyright 2009 Jay Baird <jay@mochimedia.com>

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
from twisted.internet import defer

from riakasaurus.riak_object import RiakObject

import mimetypes


class RiakBucket(object):
    """
    The ``RiakBucket`` object allows you to access and change information
    about a Riak bucket, and provides methods to create or retrieve
    objects within the bucket.
    """

    SEARCH_PRECOMMIT_HOOK = {"mod": "riak_search_kv_hook", "fun": "precommit"}

    def __init__(self, client, name):
        """
        Returns a new ``RiakBucket`` instance.

        :param client: A :class:`RiakClient <riak.client.RiakClient>` instance
        :type client: :class:`RiakClient <riak.client.RiakClient>`
        :param name: The bucket name
        :type name: string
        """
        try:
            if isinstance(name, basestring):
                name = name.encode('ascii')
        except UnicodeError:
            raise TypeError('Unicode bucket names are not supported.')

        self._client = client
        self._name = name
        self._r = None
        self._w = None
        self._dw = None
        self._rw = None
        self._pr = None
        self._pw = None
        self._encoders = {}
        self._decoders = {}

    def get_name(self):
        """
        Get the bucket name as a string.
        """
        return self._name

    def get_r(self, r=None):
        """
        Get the R-value for this bucket, if it is set, otherwise return
        the R-value for the client.

        :rtype: integer
        """
        if (r is not None):
            return r
        if (self._r is not None):
            return self._r
        return self._client.get_r()

    def set_r(self, r):
        """
        Set the R-value for this bucket. This value is used by :func:`get`
        and :func:`get_binary` operations that do not specify an R-value.

        :param r: The new R-value.
        :type r: integer
        :rtype: self
        """
        self._r = r
        return self

    def get_w(self, w=None):
        """
        Get the W-value for this bucket, if it is set, otherwise return
        the W-value for the client.

        :rtype: integer
        """
        if (w is not None):
            return w
        if (self._w is not None):
            return self._w
        return self._client.get_w()

    def set_w(self, w):
        """
        Set the W-value for this bucket. See :func:`set_r` for
        more information.

        :param w: The new W-value.
        :type w: integer
        :rtype: self
        """
        self._w = w
        return self

    def get_dw(self, dw=None):
        """
        Get the DW-value for this bucket, if it is set, otherwise return
        the DW-value for the client.

        :rtype: integer
        """
        if (dw is not None):
            return dw
        if (self._dw is not None):
            return self._dw
        return self._client.get_dw()

    def set_dw(self, dw):
        """
        Set the DW-value for this bucket. See :func:`set_r` for more
        information.

        :param dw: The new DW-value
        :type dw: integer
        :rtype: self
        """
        self._dw = dw
        return self

    def get_rw(self, rw=None):
        """
        Get the RW-value for this bucket, if it is set, otherwise return
        the RW-value for the client.

        :rtype: integer
        """
        if (rw is not None):
            return rw
        if (self._rw is not None):
            return self._rw
        return self._client.get_rw()

    def set_rw(self, rw):
        """
        Set the RW-value for this bucket. See :func:`set_r` for more
        information.

        :param rw: The new RW-value
        :type rw: integer
        :rtype: self
        """
        self._rw = rw
        return self

    def get_pr(self, pr=None):
        """
        Get the PR-value for this bucket, if it is set, otherwise return
        the PR-value for the client.

        :rtype: integer
        """
        if (pr is not None):
            return pr
        if (self._pr is not None):
            return self._pr
        return self._client.get_pr()

    def set_pr(self, pr):
        """
        Set the PR-value for this bucket. See :func:`set_r` for more
        information.

        :param pr: The new PR-value
        :type pr: integer
        :rtype: self
        """
        self._pr = pr
        return self

    def get_pw(self, pw=None):
        """
        Get the PW-value for this bucket, if it is set, otherwise return
        the PW-value for the client.

        :rtype: integer
        """
        if (pw is not None):
            return pw
        if (self._pw is not None):
            return self._pw
        return self._client.get_pw()

    def set_pw(self, pw):
        """
        Set the PW-value for this bucket. See :func:`set_r` for more
        information.

        :param pw: The new PR-value
        :type pw: integer
        :rtype: self
        """
        self._pw = pw
        return self

    def get_encoder(self, content_type):
        """
        Get the encoding function for the provided content type for this
        bucket.

        :param content_type: Content type requested
        """
        if content_type in self._encoders:
            return self._encoders[content_type]
        else:
            return self._client.get_encoder(content_type)

    def set_encoder(self, content_type, encoder):
        """
        Set the encoding function for the provided content type for this
        bucket.

        :param content_type: Content type for encoder
        :param encoder: Function to encode with - will be called with data as
                        single argument.
        """
        self._encoders[content_type] = encoder
        return self

    def get_decoder(self, content_type):
        """
        Get the decoding function for the provided content type for this
        bucket.

        :param content_type: Content type for decoder
        """
        if content_type in self._decoders:
            return self._decoders[content_type]
        else:
            return self._client.get_decoder(content_type)

    def set_decoder(self, content_type, decoder):
        """
        Set the decoding function for the provided content type for this
        bucket.

        :param content_type: Content type for decoder
        :param decoder: Function to decode with - will be called with string
        """
        self._decoders[content_type] = decoder
        return self

    def new(self, key=None, data=None, content_type='application/json'):
        """
        Create a new :class:`RiakObject <riak.riak_object.RiakObject>` that
        will be stored as JSON. A shortcut for manually instantiating a
        :class:`RiakObject <riak.riak_object.RiakObject>`.

        :param key: Name of the key. Leaving this to be None (default) will
                    make Riak generate the key on store.
        :type key: string
        :param data: The data to store.
        :type data: object
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        """
        try:
            if isinstance(data, basestring):
                data = data.encode('ascii')
        except UnicodeError:
            raise TypeError('Unicode data values are not supported.')

        obj = RiakObject(self._client, self, key)
        obj.set_data(data)
        obj.set_content_type(content_type)
        obj._encode_data = True
        return obj

    def new_binary(self, key, data, content_type='application/octet-stream'):
        """
        Create a new :class:`RiakObject <riak.riak_object.RiakObject>` that
        will be stored as plain text/binary.
        A shortcut for manually instantiating a
        :class:`RiakObject <riak.riak_object.RiakObject>`.

        :param key: Name of the key.
        :type key: string
        :param data: The data to store.
        :type data: object
        :param content_type: The content type of the object.
        :type content_type: string
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        """
        obj = RiakObject(self._client, self, key)
        obj.set_data(data)
        obj.set_content_type(content_type)
        obj._encode_data = False
        return obj

    def get(self, key, r=None, pr=None):
        """
        Retrieve a JSON-encoded object from Riak.

        :param key: Name of the key.
        :type key: string
        :param r: R-Value of the request (defaults to bucket's R)
        :type r: integer
        :param pr: PR-Value of the request (defaults to bucket's PR)
        :type pr: integer
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        """
        obj = RiakObject(self._client, self, key)
        obj._encode_data = True
        r = self.get_r(r)
        pr = self.get_pr(pr)
        return obj.reload(r=r, pr=pr)

    def head(self, key, r=None, pr=None):
        """
        Retrieve a JSON-encoded object from Riak.

        :param key: Name of the key.
        :type key: string
        :param r: R-Value of the request (defaults to bucket's R)
        :type r: integer
        :param pr: PR-Value of the request (defaults to bucket's PR)
        :type pr: integer
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        """
        obj = RiakObject(self._client, self, key)
        obj._encode_data = True
        r = self.get_r(r)
        pr = self.get_pr(pr)
        return obj.head(r=r, pr=pr)

    def get_binary(self, key, r=None, pr=None):
        """
        Retrieve a binary/string object from Riak.

        :param key: Name of the key.
        :type key: string
        :param r: R-Value of the request (defaults to bucket's R)
        :type r: integer
        :param pr: PR-Value of the request (defaults to bucket's PR)
        :type pr: integer
        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
        """
        obj = RiakObject(self._client, self, key)
        obj._encode_data = False
        r = self.get_r(r)
        pr = self.get_pr(pr)
        return obj.reload(r=r, pr=pr)

    def set_n_val(self, nval):
        """
        Set the N-value for this bucket, which is the number of replicas
        that will be written of each object in the bucket.

        .. warning::

           Set this once before you write any data to the bucket, and never
           change it again, otherwise unpredictable things could happen.
           This should only be used if you know what you are doing.

        :param nval: The new N-Val.
        :type nval: integer
        """
        return self.set_property('n_val', nval)

    def get_n_val(self):
        """
        Retrieve the N-value for this bucket.

        :rtype: integer
        """
        return self.get_property('n_val')

    def set_default_r_val(self, rval):
        return self.set_property('r', rval)

    def get_default_r_val(self):
        return self.get_property('r')

    def set_default_w_val(self, wval):
        return self.set_property('w', wval)

    def get_default_w_val(self):
        return self.get_property('w')

    def set_default_dw_val(self, dwval):
        return self.set_property('dw', dwval)

    def get_default_dw_val(self):
        return self.get_property('dw')

    def set_default_rw_val(self, rwval):
        return self.set_property('rw', rwval)

    def get_default_rw_val(self):
        return self.get_property('rw')

    def set_allow_multiples(self, bool):
        """
        If set to True, then writes with conflicting data will be stored
        and returned to the client. This situation can be detected by
        calling has_siblings() and get_siblings().

        .. warning::

           This should only be used if you know what you are doing, as it can
           lead to unexpected results.

        :param bool: True to store and return conflicting writes.
        :type bool: boolean
        """
        return self.set_property('allow_mult', bool)

    def get_allow_multiples(self):
        """
        Retrieve the 'allow multiples' setting.

        :rtype: Boolean
        """
        return self.get_bool_property('allow_mult')

    def set_property(self, key, value):
        """
        Set a bucket property.

        .. warning::

           This should only be used if you know what you are doing.

        :param key: Property to set.
        :type key: string
        :param value: Property value.
        :type value: mixed
        """
        return self.set_properties({key: value})

    @defer.inlineCallbacks
    def get_bool_property(self, key):
        """
        Get a boolean bucket property.  Converts to a ``True`` or ``False``
        value.

        :param key: Property to set.
        :type key: string
        """
        prop = yield self.get_property(key)
        if prop == True or prop > 0:
            defer.returnValue(True)
        else:
            defer.returnValue(False)

    @defer.inlineCallbacks
    def get_property(self, key):
        """
        Retrieve a bucket property.

        :param key: The property to retrieve.
        :type key: string
        :rtype: mixed
        """
        props = yield self.get_properties()
        if (key in props.keys()):
            defer.returnValue(props[key])
        else:
            defer.returnValue(None)

    def set_properties(self, props):
        """
        Set multiple bucket properties in one call.

        .. warning::

           This should only be used if you know what you are doing.

        :param props: An associative array of key:value.
        :type props: array - deferred
        """
        return self._client.transport.set_bucket_props(self, props)

    def get_properties(self):
        """
        Retrieve an associative array of all bucket properties.

        :rtype: array - deferred
        """
        return self._client.transport.get_bucket_props(self)

    def reset_properties(self):
        """
        Reset all bucket properties to defaults.

        :rtype: None - deferred
        """
        return self._client.transport.reset_bucket_props(self)

    def get_keys(self):
        """
        Return all keys within the bucket.

        .. warning::

           At current, this is a very expensive operation. Use with caution.
        """
        return self._client.transport.get_keys(self)

    def new_binary_from_file(self, key, filename):
        """
        Create a new Riak object in the bucket, using the content of the
        specified file.
        """
        binary_data = open(filename, "rb").read()
        mimetype, encoding = mimetypes.guess_type(filename)
        if not mimetype:
            mimetype = 'application/octet-stream'
        return self.new_binary(key, binary_data, mimetype)

    @defer.inlineCallbacks
    def search_enabled(self):
        """
        Returns True if the search precommit hook is enabled for this bucket.
        """
        pch = yield self.get_property("precommit")
        defer.returnValue(self.SEARCH_PRECOMMIT_HOOK in (pch or []))

    @defer.inlineCallbacks
    def enable_search(self):
        """
        Enable search for this bucket by installing the precommit hook to
        index objects in it.
        Returns deferred
        """
        pch = yield self.get_property("precommit")
        precommit_hooks = pch or []
        if self.SEARCH_PRECOMMIT_HOOK not in precommit_hooks:
            yield self.set_properties({"precommit":
                precommit_hooks + [self.SEARCH_PRECOMMIT_HOOK]})

        defer.returnValue(True)

    @defer.inlineCallbacks
    def disable_search(self):
        """
        Disable search for this bucket by removing the precommit hook to
        index objects in it.
        """
        pch = yield self.get_property("precommit")
        precommit_hooks = pch or []
        if self.SEARCH_PRECOMMIT_HOOK in precommit_hooks:
            precommit_hooks.remove(self.SEARCH_PRECOMMIT_HOOK)
            yield self.set_properties({"precommit": precommit_hooks})

        defer.returnValue(True)

    def search(self, query, **params):
        """
        Queries a search index over objects in this bucket/index.
        """
        return self._client.solr().search(self._name, query, **params)

    def get_index(self, index, startkey, endkey=None):
        """
        Queries a secondary index over objects in this bucket, returning keys.
        """
        return self._client.transport.get_index(
            self._name, index, startkey, endkey)

    def list_keys(self):
        """ Same as get_keys - for txRiak compat """
        return self.get_keys()

    @defer.inlineCallbacks
    def purge_keys(self):
        """
        Purge all keys from the bucket. Specific to Riakasaurus

        :returns: None

        This is a convenience function that lists all of the keys
        in the bucket and then deletes them.

        NB: This is a VERY resource-intensive operation, and is
            IRREVERSIBLE. Be careful.
        """

        # Get the current key list
        keys = yield self.get_keys()

        # Major key-killing action
        for key in keys:
            obj = yield self.get_binary(key)
            yield obj.delete()



class RequestTimeout(Exception):
    """
        Raised to notify that an operation exceeded its timeout.
    """


class ConnectTimeout(Exception):
    """
        Raised to notify that a connection timed out
    """


class RiakPBCException(Exception):
    """Generic PBC exception"""
    pass


class RiakError(Exception):
    """Generic Riak error"""
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)



from zope.interface import implements

from twisted.protocols.basic import LineReceiver
LineReceiver.MAX_LENGTH = 1024 * 1024 * 64

from twisted.internet import defer, reactor, protocol
from twisted.python import log
import logging

from distutils.version import LooseVersion

import time

# MD_ resources
from riakasaurus.metadata import *

from riakasaurus.riak_index_entry import RiakIndexEntry
from riakasaurus.mapreduce import RiakLink
from riakasaurus import exceptions

# protobuf
from riakasaurus.transport import transport, pbc

LOGLEVEL_DEBUG = 1
LOGLEVEL_TRANSPORT = 2
LOGLEVEL_TRANSPORT_VERBOSE = 4


class StatefulTransport(object):
    def __init__(self):
        self.__transport = None
        self.__state = 'idle'
        self.__created = time.time()
        self.__used = time.time()

    def __repr__(self):
        return '<StatefulTransport idle=%.2fs state=\'%s\' transport=%s>' % (
            time.time() - self.__used, self.__state, self.__transport)

    def __enter__(self):
        return self.getTransport()

    def __exit__(self, exc_type, exc_value, traceback):
        self.setIdle()

    def isActive(self):
        return self.__state == 'active'

    def setActive(self):
        self.__state = 'active'
        self.__used = time.time()

    def isIdle(self):
        return self.__state == 'idle'

    def setIdle(self):
        self.__state = 'idle'
        self.__used = time.time()

    def setTransport(self, transport):
        self.__transport = transport

    def getTransport(self):
        return self.__transport

    def age(self):
        return time.time() - self.__used

    def isDisconnected(self):
        transport = self.getTransport()
        return transport and transport.isDisconnected()


class PBCTransport(transport.FeatureDetection):
    """ Protocoll buffer transport for Riak """

    implements(transport.ITransport)

    debug = 0
    logToLevel = logging.INFO
    MAX_TRANSPORTS = 50
    MAX_IDLETIME = 5 * 60     # in seconds
    # how often (in seconds) the garbage collection should run
    # XXX Why the hell do we even have to override GC?
    GC_TIME = 120

    def __init__(self, client):
        self._prefix = client._prefix
        self.host = client._host
        self.port = client._port
        self.client = client
        self._client_id = None
        self._transports = []    # list of transports, empty on start
        self._gc = reactor.callLater(self.GC_TIME, self._garbageCollect)
        self.timeout = client.request_timeout

    def setTimeout(self, t):
        self.timeout = t

    @defer.inlineCallbacks
    def _getFreeTransport(self):
        foundOne = False

        # Discard disconnected transports.
        self._transports = [x for x in self._transports if not x.isDisconnected()]

        for stp in self._transports:
            if stp.isIdle():
                stp.setActive()
                stp.getTransport().setTimeout(self.timeout)
                foundOne = True
                if self.debug & LOGLEVEL_TRANSPORT_VERBOSE:
                    log.msg("[%s] aquired idle transport[%d]: %s" % (
                            self.__class__.__name__,
                            len(self._transports), stp
                        ), logLevel=self.logToLevel)
                defer.returnValue(stp)
        if not foundOne:
            if len(self._transports) >= self.MAX_TRANSPORTS:
                raise Exception("too many transports, aborting")

            # nothin free, create a new protocol instance, append
            # it to self._transports and return it

            # insert a placeholder into self._transports to avoid race
            # conditions with the MAX_TRANSPORTS check above.
            stp = StatefulTransport()
            stp.setActive()
            idx = len(self._transports)
            self._transports.append(stp)

            # create the transport and use it to configure the placeholder.
            try:
                transport = yield pbc.RiakPBCClient().connect(self.host, self.port)
                stp.setTransport(transport)
                if self.timeout:
                    transport.setTimeout(self.timeout)
                if self.debug & LOGLEVEL_TRANSPORT:
                    log.msg("[%s] allocate new transport[%d]: %s" % (
                            self.__class__.__name__, idx, stp
                        ), logLevel=self.logToLevel)
                defer.returnValue(stp)
            except Exception:
                self._transports.remove(stp)
                raise


    @defer.inlineCallbacks
    def _garbageCollect(self):
        self._gc = reactor.callLater(self.GC_TIME, self._garbageCollect)
        for idx, stp in enumerate(self._transports):
            if (stp.isIdle() and stp.age() > self.MAX_IDLETIME):
                yield stp.getTransport().quit()
                if self.debug & LOGLEVEL_TRANSPORT:
                    log.msg("[%s] expire idle transport[%d] %s" % (
                            self.__class__.__name__,
                            idx,
                            stp
                        ), logLevel=self.logToLevel)

                    log.msg("[%s] %s" % (
                            self.__class__.__name__,
                            self._transports
                        ), logLevel=self.logToLevel)

                self._transports.remove(stp)
            elif self.timeout and stp.isActive() and stp.age() > self.timeout:
                yield stp.getTransport().quit()
                if self.debug & LOGLEVEL_TRANSPORT:
                    log.msg("[%s] expire timeouted transport[%d] %s" % (
                            self.__class__.__name__,
                            idx,
                            stp
                        ), logLevel=self.logToLevel)

                    log.msg("[%s] %s" % (
                            self.__class__.__name__,
                            self._transports
                        ), logLevel=self.logToLevel)

                self._transports.remove(stp)

    @defer.inlineCallbacks
    def quit(self):
        self._gc.cancel()      # cancel the garbage collector

        for stp in self._transports:
            if self.debug & LOGLEVEL_DEBUG:
                log.msg("[%s] transport[%d].quit() %s" % (
                        self.__class__.__name__,
                        len(self._transports),
                        stp
                    ), logLevel=self.logToLevel)

            yield (stp.getTransport() and stp.getTransport().quit())

    def __del__(self):
        """on shutdown, close all transports"""
        self.quit()

    def put(self, robj, w=None, dw=None, pw=None, return_body=True,
            if_none_match=False):

        ret = self.__put(robj, w, dw, pw,
                         return_body=return_body, if_none_match=if_none_match)

        if return_body:
            return ret
        else:
            return None

    def put_new(self, robj, w=None, dw=None, pw=None, return_body=True,
                if_none_match=False):
        ret = self.__put(robj, w, dw, pw,
                         return_body=return_body, if_none_match=if_none_match)
        if return_body:
            return ret
        else:
            return (ret[0], None, None)

    @defer.inlineCallbacks
    def __put(self, robj, w=None, dw=None, pw=None, return_body=True,
              if_none_match=False):
        # std kwargs
        kwargs = {
                'w': w,
                'dw': dw,
                'pw': pw,
                'return_body': return_body,
                'if_none_match': if_none_match
            }
        # vclock
        vclock = robj.vclock() or None

        payload = {
                'value': robj.get_encoded_data(),
                'content_type': robj.get_content_type(),
            }

        # links
        links = robj.get_links()
        if links:
            payload['links'] = []
            for l in links:
                payload['links'].append(
                    (l.get_bucket(), l.get_key(), l.get_tag())
                )

        # usermeta
        if robj.get_usermeta():
            payload['usermeta'] = []
            for key, value in robj.get_usermeta().iteritems():
                payload['usermeta'].append((key, value))

        # indexes
        if robj.get_indexes():
            payload['indexes'] = []
            for index in robj.get_indexes():
                payload['indexes'].append(
                    (index.get_field(), index.get_value())
                )

        # aquire transport, fire, release
        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.put(robj.get_bucket().get_name(),
                                      robj.get_key(),
                                      payload,
                                      vclock,
                                      **kwargs
                                      )
        defer.returnValue(self.parseRpbGetResp(ret))

    @defer.inlineCallbacks
    def get(self, robj, r=None, pr=None, vtag=None):

        # ***FIXME*** whats vtag for? ignored for now

        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.get(robj.get_bucket().get_name(),
                                      robj.get_key(),
                                      r=r,
                                      pr=pr)

        defer.returnValue(self.parseRpbGetResp(ret))

    @defer.inlineCallbacks
    def head(self, robj, r=None, pr=None, vtag=None):
        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.get(robj.get_bucket().get_name(),
                                      robj.get_key(),
                                      r=r,
                                      pr=pr,
                                      head=True)

        defer.returnValue(self.parseRpbGetResp(ret))

    @defer.inlineCallbacks
    def delete(self, robj, rw=None, r=None, w=None, dw=None, pr=None, pw=None):
        """
        Delete an object.
        """
        # We could detect quorum_controls here but HTTP ignores
        # unknown flags/params.
        kwargs = {'rw': rw, 'r': r, 'w': w, 'dw': dw, 'pr': pr, 'pw': pw}
        headers = {}

        ts = yield self.tombstone_vclocks()
        if ts and robj.vclock() is not None:
            kwargs['vclock'] = robj.vclock()

        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.delete(robj.get_bucket().get_name(),
                                         robj.get_key(),
                                         **kwargs
                                         )

        defer.returnValue(ret)

    @defer.inlineCallbacks
    def get_buckets(self):
        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.getBuckets()
        defer.returnValue([x for x in ret.buckets])

    @defer.inlineCallbacks
    def server_version(self):
        if not self._s_version:
            self._s_version = yield self._server_version()

        defer.returnValue(LooseVersion(self._s_version))

    @defer.inlineCallbacks
    def _server_version(self):
        with (yield self._getFreeTransport()) as transport:
            stats = yield transport.getServerInfo()

        if stats is not None:
            if self.debug % LOGLEVEL_DEBUG:
                log.msg("[%s] fetched server version: %s" % (
                    self.__class__.__name__,
                    stats.server_version
                ), logLevel=self.logToLevel)
            defer.returnValue(stats.server_version)
        else:
            defer.returnValue("0.14.0")

    @defer.inlineCallbacks
    def ping(self):
        """
        Check server is alive
        """
        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.ping()
        defer.returnValue(ret == True)

    @defer.inlineCallbacks
    def set_bucket_props(self, bucket, props):
        """
        Set bucket properties
        """
        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.setBucketProperties(bucket.get_name(), **props)
        defer.returnValue(ret == True)

    @defer.inlineCallbacks
    def get_bucket_props(self, bucket):
        """
        get bucket properties
        """
        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.getBucketProperties(bucket.get_name())
        defer.returnValue({
            'n_val': ret.props.n_val,
            'allow_mult': ret.props.allow_mult
        })

    @defer.inlineCallbacks
    def get_keys(self, bucket):
        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.getKeys(bucket.get_name())
        defer.returnValue(ret)

    @defer.inlineCallbacks
    def get_index(self, bucket, index, startkey, endkey=None):
        with (yield self._getFreeTransport()) as transport:
            ret = yield transport.get_index(bucket, index, startkey, endkey=endkey)
        defer.returnValue(ret)

    def parseRpbGetResp(self, res):
        """
        adaptor for a RpbGetResp message
        message RpbGetResp {
           repeated RpbContent content = 1;
           optional bytes vclock = 2;
           optional bool unchanged = 3;
        }
        """
        if res == True:         # empty response
            return None
        vclock = res.vclock
        resList = []
        for content in res.content:
            # iterate over RpbContent field
            metadata = {MD_USERMETA: {}, MD_INDEX: []}
            data = content.value
            if content.HasField('content_type'):
                metadata[MD_CTYPE] = content.content_type

            if content.HasField('charset'):
                metadata[MD_CHARSET] = content.charset

            if content.HasField('content_encoding'):
                metadata[MD_ENCODING] = content.content_encoding

            if content.HasField('vtag'):
                metadata[MD_VTAG] = content.vtag

            if content.HasField('last_mod'):
                metadata[MD_LASTMOD] = content.last_mod

            if content.HasField('deleted'):
                metadata[MD_DELETED] = content.deleted

            if len(content.links):
                metadata[MD_LINKS] = []
                for l in content.links:
                    metadata[MD_LINKS].append(
                        RiakLink(l.bucket, l.key, l.tag)
                    )

            if len(content.usermeta):
                metadata[MD_USERMETA] = {}
                for md in content.usermeta:
                    metadata[MD_USERMETA][md.key] = md.value

            if len(content.indexes):
                metadata[MD_INDEX] = []
                for ie in content.indexes:
                    metadata[MD_INDEX].append(RiakIndexEntry(ie.key, ie.value))
            resList.append((metadata, data))
        return vclock, resList

    def decodeJson(self, s):
        return self.client.get_decoder('application/json')(s)

    def encodeJson(self, s):
        return self.client.get_encoder('application/json')(s)

    # def deferred_sleep(self,secs):
    #     """
    #     fake deferred sleep

    #     @param secs: time to sleep
    #     @type secs: float
    #     """
    #     d = defer.Deferred()
    #     reactor.callLater(secs, d.callback, None)
    #     return d



"""
Copyright 2010 Rusty Klophaus <rusty@basho.com>
Copyright 2010 Justin Sheehy <justin@basho.com>
Copyright 2009 Jay Baird <jay@mochimedia.com>

This file is provided to you under the Apache License,
Version 2.0 (the "License"); you may not use this file
except in compliance with the License.  You may obtain
a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""


class RiakIndexEntry:
    def __init__(self, field, value):
        self._field = field
        self._value = str(value)

    def get_field(self):
        return self._field

    def get_value(self):
        return self._value

    def __str__(self):
        return "RiakIndexEntry(field = '%s', value='%s')" % (
            self._field, self._value)

    def __eq__(self, other):
        if not isinstance(other, RiakIndexEntry):
            return False

        return \
            self.get_field() == other.get_field() and \
            self.get_value() == other.get_value()

    def __cmp__(self, other):
        if other == None:
            raise TypeError("RiakIndexEntry cannot be compared to None")

        if not isinstance(other, RiakIndexEntry):
            raise TypeError("RiakIndexEntry cannot be compared to %s" %
                other.__class__.__name__)

        if self.get_field() < other.get_field():
            return -1

        if self.get_field() > other.get_field():
            return 1

        if self.get_value() < other.get_value():
            return -1

        if self.get_value() > other.get_value():
            return 1

        return 0



#!/usr/bin/env python
"""
riakasaurus trial test file.
riakasaurus _must_ be on your PYTHONPATH

"""

import json
import random
from twisted.trial import unittest
from twisted.python import log
from twisted.internet import defer

VERBOSE = False

from riakasaurus import riak

# uncomment to activate logging
# import sys
# log.startLogging(sys.stderr)

RIAK_CLIENT_ID = 'TEST'
BUCKET_PREFIX = 'riakasaurus.tests.'

JAVASCRIPT_SUM = """
function(v) {
  x = v.reduce(function(a,b){ return a + b }, 0);
  return [x];
}
"""


def randint():
    """Generate nice random int for our test."""
    return random.randint(1, 999999)


class Tests(unittest.TestCase):
    """
    trial unit tests.
    """

    test_keys = ['foo', 'foo1', 'foo2', 'foo3', 'bar', 'baz', 'ba_foo1',
                 'blue_foo1']

    @defer.inlineCallbacks
    def setUp(self):
        self.client = riak.RiakClient(client_id=RIAK_CLIENT_ID)
        self.bucket_name = BUCKET_PREFIX + self.id().rsplit('.', 1)[-1]
        self.bucket = self.client.bucket(self.bucket_name)
        yield self.bucket.enable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.bucket.disable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def test_secondary_index_mapred(self):
        log.msg("*** secondary_index")
        yield self.bucket.enable_search()

        obj = self.bucket.new('foo1', {'field1': 'val1', 'field2': 1001})
        obj.add_index('field1_bin', 'val1')
        obj.add_index('field2_int', 1001)
        yield obj.store()

        obj = self.bucket.new('foo2', {'field1': 'val2', 'field2': 1003})
        obj.add_index('field1_bin', 'val2')
        obj.add_index('field2_int', 1003)
        yield obj.store()

        results = yield self.client.index(self.bucket_name,
                                          'field1_bin', 'val2').run()

        r1 = yield results[0].get()

        self.assertEqual(r1.get_key(), u'foo2')

        log.msg("done secondary_index")

    @defer.inlineCallbacks
    def test_get_index(self):
        log.msg("*** secondary_index")
        yield self.bucket.enable_search()

        obj = self.bucket.new('foo1', {'field1': 'val1', 'field2': 1001})
        obj.add_index('field1_bin', 'val1')
        obj.add_index('field2_int', 1001)
        yield obj.store()

        obj = self.bucket.new('foo2', {'field1': 'val2', 'field2': 1003})
        obj.add_index('field1_bin', 'val2')
        obj.add_index('field2_int', 1003)
        yield obj.store()

        results = yield self.bucket.get_index('field2_int', 1,
                                          2000)

        self.assertEqual(sorted(results),
                         ['foo1', 'foo2'])

        log.msg("done secondary_index")



#!/usr/bin/env python
"""
tests for RiakPBCClient, trial
"""

from twisted.trial import unittest
from twisted.internet import defer
from twisted.python import log
from twisted.internet.error import ConnectionRefusedError

from riakasaurus.transport import pbc

from riakasaurus import riak, transport

VERBOSE = False
# uncomment to activate logging
if VERBOSE:
    import sys
    log.startLogging(sys.stderr)


class Test_PBCClient(unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.client = yield pbc.RiakPBCClient().connect('127.0.0.1', 8087)
        self.client.debug = 0

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.client.quit()

    @defer.inlineCallbacks
    def test_ping(self):
        log.msg("*** testing ping")
        res = yield self.client.ping()
        self.assertTrue(res)
        log.msg("done testing ping")

    @defer.inlineCallbacks
    def test_setgetClientId(self):
        log.msg("*** testing getClientId")
        yield self.client.setClientId('MyClientId')
        res = yield self.client.getClientId()
        self.assertEqual(res.client_id, 'MyClientId')
        log.msg("done testing getClientId")

    @defer.inlineCallbacks
    def test_getServerInfo(self):
        log.msg("*** testing getServerInfo")
        info = yield self.client.getServerInfo()
        self.assertTrue(isinstance(info, pbc.RpbGetServerInfoResp))
        log.msg("done testing getServerInfo")

    @defer.inlineCallbacks
    def test_put(self):
        log.msg("*** testing put")
        put = yield self.client.put('bucket', 'key', 'foo')
        self.assertTrue(isinstance(put, pbc.RpbPutResp))

        log.msg("done testing put")

    @defer.inlineCallbacks
    def test_get(self):
        log.msg("*** testing get")
        put = yield self.client.put('bucket', 'key', 'foo')
        self.assertTrue(isinstance(put, pbc.RpbPutResp))

        result = yield self.client.get('bucket', 'key')
        self.assertTrue(isinstance(result, pbc.RpbGetResp))
        self.assertEqual(result.content[0].value, 'foo')

        log.msg("done testing get")

    @defer.inlineCallbacks
    def test_update(self):
        log.msg("*** testing update")
        # make sure "foo" is in
        put = yield self.client.put('bucket', 'key', 'foo')
        self.assertTrue(isinstance(put, pbc.RpbPutResp))

        # retrieve it
        result = yield self.client.get('bucket', 'key')
        self.assertTrue(isinstance(result, pbc.RpbGetResp))
        self.assertEqual(result.content[0].value, 'foo')

        result2 = yield self.client.put('bucket', 'key', 'bla', result.vclock)
        self.assertTrue(isinstance(result2, pbc.RpbPutResp))

        # retrieve updated
        result = yield self.client.get('bucket', 'key')
        self.assertTrue(isinstance(result, pbc.RpbGetResp))
        self.assertEqual(result.content[0].value, 'bla')

        log.msg("done testing update")

    @defer.inlineCallbacks
    def test_delete(self):
        log.msg("*** testing delete")

        # make sure "foo" is in
        put = yield self.client.put('bucket', 'delete_key', 'foo')
        self.assertTrue(isinstance(put, pbc.RpbPutResp))

        # remove it
        result = yield self.client.delete('bucket', 'delete_key')
        self.assertEqual(result, True)

        # try to fetch it
        result = yield self.client.get('bucket', 'delete_key')
        self.assertTrue(isinstance(result, pbc.RpbGetResp))
        self.assertEqual(len(result.content), 0)
        log.msg("done testing delete")

    @defer.inlineCallbacks
    def test_getKeys(self):
        log.msg("*** testing getKeys")

        # get existing keys
        res = yield self.client.getKeys('bucket')
        self.assertTrue(isinstance(res, list))

        log.msg("done testing getKeys")

    @defer.inlineCallbacks
    def test_getBuckets(self):
        log.msg("*** testing getBuckets")

        # make sure "foo" is in
        put = yield self.client.put('bucket', 'key', 'foo')
        self.assertTrue(isinstance(put, pbc.RpbPutResp))

        # get existing buckets message
        res = yield self.client.getBuckets()
        buckets = [b for b in res.buckets]
        self.assertTrue('bucket' in buckets)
        log.msg("done testing getBuckets")

    @defer.inlineCallbacks
    def test_setGetBucketProperties(self):
        log.msg("*** testing setGetBucketProperties")

        # make sure "foo" is in
        put = yield self.client.put('anotherbucket', 'key', 'foo')
        self.assertTrue(isinstance(put, pbc.RpbPutResp))

        # get existing buckets message
        res = yield self.client.setBucketProperties('anotherbucket', n_val=2,
            allow_mult=True)
        self.assertTrue(res == True)

        res = yield self.client.getBucketProperties('anotherbucket')
        self.assertEqual(res.props.n_val, 2)

        yield self.client.delete('anotherbucket', 'key')

        log.msg("done testing setGetBucketProperties")

    @defer.inlineCallbacks
    def test_links(self):
        log.msg("*** testing links")

        # retrieve it
        head = yield self.client.get('bucket', 'key', head=True)

        # make sure "foo" is in
        content = {
            'value': 'foo',
            'content_type': 'text/text',
            'content_encoding': 'UTF8',
            'links': [
                ('bucket', 'subkey', 'tag'),
                ('bucket', 'nochnkey', ''),
            ]
        }
        put = yield self.client.put('bucket', 'key', content,
            vclock=head.vclock)

        self.assertTrue(isinstance(put, pbc.RpbPutResp))

        # retrieve it
        result = yield self.client.get('bucket', 'key')
        self.assertTrue(isinstance(result, pbc.RpbGetResp))

        self.assertEqual(result.content[0].value, 'foo')
        self.assertEqual(result.content[0].content_type, 'text/text')
        self.assertEqual(result.content[0].content_encoding, 'UTF8')
        self.assertTrue(len(result.content[0].links) == 2)

        log.msg("done testing links")

    @defer.inlineCallbacks
    def test_connection_refused(self):
        try:
            # Connection to a host that does not exist.
            yield pbc.RiakPBCClient().connect('127.0.0.17', 8087)
            assert False, 'ConnectionRefusedError not raised as expected.'
        except ConnectionRefusedError:
            pass



from twisted.protocols.basic import Int32StringReceiver
from twisted.internet.protocol import ClientFactory
from twisted.internet.defer import Deferred
from twisted.internet import defer, reactor
from twisted.python.failure import Failure

from struct import pack, unpack

from pprint import pformat

# generated code from *.proto message definitions
from riakasaurus.transport.pbc import riak_kv_pb2, riak_pb2
from riakasaurus import exceptions

## Protocol codes
MSG_CODE_ERROR_RESP = 0
MSG_CODE_PING_REQ = 1
MSG_CODE_PING_RESP = 2
MSG_CODE_GET_CLIENT_ID_REQ = 3
MSG_CODE_GET_CLIENT_ID_RESP = 4
MSG_CODE_SET_CLIENT_ID_REQ = 5
MSG_CODE_SET_CLIENT_ID_RESP = 6
MSG_CODE_GET_SERVER_INFO_REQ = 7
MSG_CODE_GET_SERVER_INFO_RESP = 8
MSG_CODE_GET_REQ = 9
MSG_CODE_GET_RESP = 10
MSG_CODE_PUT_REQ = 11
MSG_CODE_PUT_RESP = 12
MSG_CODE_DEL_REQ = 13
MSG_CODE_DEL_RESP = 14
MSG_CODE_LIST_BUCKETS_REQ = 15
MSG_CODE_LIST_BUCKETS_RESP = 16
MSG_CODE_LIST_KEYS_REQ = 17
MSG_CODE_LIST_KEYS_RESP = 18
MSG_CODE_GET_BUCKET_REQ = 19
MSG_CODE_GET_BUCKET_RESP = 20
MSG_CODE_SET_BUCKET_REQ = 21
MSG_CODE_SET_BUCKET_RESP = 22
MSG_CODE_MAPRED_REQ = 23
MSG_CODE_MAPRED_RESP = 24
MSG_CODE_INDEX_REQ = 25
MSG_CODE_INDEX_RESP = 26
MSG_CODE_SEARCH_QUERY_REQ = 27
MSG_CODE_SEARCH_QUERY_RESP = 28

## Inject PBC classes

RpbGetClientIdResp = riak_kv_pb2.RpbGetClientIdResp
RpbSetClientIdReq = riak_kv_pb2.RpbSetClientIdReq
RpbGetReq = riak_kv_pb2.RpbGetReq
RpbGetResp = riak_kv_pb2.RpbGetResp
RpbPutReq = riak_kv_pb2.RpbPutReq
RpbPutResp = riak_kv_pb2.RpbPutResp
RpbDelReq = riak_kv_pb2.RpbDelReq
RpbListBucketsResp = riak_kv_pb2.RpbListBucketsResp
RpbListKeysReq = riak_kv_pb2.RpbListKeysReq
RpbListKeysResp = riak_kv_pb2.RpbListKeysResp
RpbGetBucketReq = riak_kv_pb2.RpbGetBucketReq
RpbGetBucketResp = riak_kv_pb2.RpbGetBucketResp
RpbSetBucketReq = riak_kv_pb2.RpbSetBucketReq
RpbMapRedReq = riak_kv_pb2.RpbMapRedReq
RpbMapRedResp = riak_kv_pb2.RpbMapRedResp
RpbIndexReq = riak_kv_pb2.RpbIndexReq
RpbIndexResp = riak_kv_pb2.RpbIndexResp
RpbContent = riak_kv_pb2.RpbContent
RpbLink = riak_kv_pb2.RpbLink
RpbBucketProps = riak_kv_pb2.RpbBucketProps

RpbErrorResp = riak_pb2.RpbErrorResp
RpbGetServerInfoResp = riak_pb2.RpbGetServerInfoResp
RpbPair = riak_pb2.RpbPair


def toHex(s):
    lst = []
    for ch in s:
        hv = hex(ord(ch)).replace('0x', '')
        if len(hv) == 1:
            hv = '0' + hv
        lst.append(hv + ' ')

    return reduce(lambda x, y: x + y, lst)


class RiakPBC(Int32StringReceiver):

    MAX_LENGTH = 9999999

    riakResponses = {
        MSG_CODE_ERROR_RESP: RpbErrorResp,
        MSG_CODE_GET_CLIENT_ID_RESP: RpbGetClientIdResp,
        MSG_CODE_GET_SERVER_INFO_RESP: RpbGetServerInfoResp,
        MSG_CODE_GET_RESP: RpbGetResp,
        MSG_CODE_PUT_RESP: RpbPutResp,
        MSG_CODE_LIST_KEYS_RESP: RpbListKeysResp,
        MSG_CODE_LIST_BUCKETS_RESP: RpbListBucketsResp,
        MSG_CODE_GET_BUCKET_RESP: RpbGetBucketResp,
        MSG_CODE_GET_SERVER_INFO_RESP: RpbGetServerInfoResp,
        MSG_CODE_INDEX_RESP: RpbIndexResp,
    }

    PBMessageTypes = {
        0: 'ERROR_RESP',
        1: 'PING_REQ',
        2: 'PING_RESP',
        3: 'GET_CLIENT_ID_REQ',
        4: 'GET_CLIENT_ID_RESP',
        5: 'SET_CLIENT_ID_REQ',
        6: 'SET_CLIENT_ID_RESP',
        7: 'GET_SERVER_INFO_REQ',
        8: 'GET_SERVER_INFO_RESP',
        9: 'GET_REQ',
        10: 'GET_RESP',
        11: 'PUT_REQ',
        12: 'PUT_RESP',
        13: 'DEL_REQ',
        14: 'DEL_RESP',
        15: 'LIST_BUCKETS_REQ',
        16: 'LIST_BUCKETS_RESP',
        17: 'LIST_KEYS_REQ',
        18: 'LIST_KEYS_RESP',
        19: 'GET_BUCKET_REQ',
        20: 'GET_BUCKET_RESP',
        21: 'SET_BUCKET_REQ',
        22: 'SET_BUCKET_RESP',
        23: 'MAPRED_REQ',
        24: 'MAPRED_RESP',
        25: 'INDEX_REQ',
        26: 'INDEX_RESP',
        27: 'SEARCH_QUERY_REQ',
        28: 'SEARCH_QUERY_RESP',
    }

    nonMessages = (
        MSG_CODE_PING_RESP,
        MSG_CODE_DEL_RESP,
        MSG_CODE_SET_BUCKET_RESP,
        MSG_CODE_SET_CLIENT_ID_RESP
    )

    rwNums = {
        'one': 4294967295 - 1,
        'quorum': 4294967295 - 2,
        'all': 4294967295 - 3,
        'default': 4294967295 - 4,
    }

    timeout = None
    timeoutd = None
    disconnected = False
    debug = 0

    # ------------------------------------------------------------------
    # Server Operations .. setClientId, getClientId, getServerInfo, ping
    # ------------------------------------------------------------------
    def setClientId(self, clientId):
        code = pack('B', MSG_CODE_SET_CLIENT_ID_REQ)
        request = RpbSetClientIdReq()
        request.client_id = clientId
        return self.__send(code, request)

    def getClientId(self):
        code = pack('B', MSG_CODE_GET_CLIENT_ID_REQ)
        return self.__send(code)

    def getServerInfo(self):
        code = pack('B', MSG_CODE_GET_SERVER_INFO_REQ)
        return self.__send(code)

    def ping(self):
        code = pack('B', MSG_CODE_PING_REQ)
        return self.__send(code)

    # ------------------------------------------------------------------
    # Object/Key Operations .. get(fetch), put(store), delete
    # ------------------------------------------------------------------
    def get(self, bucket, key, **kwargs):
        code = pack('B', MSG_CODE_GET_REQ)
        request = RpbGetReq()
        request.bucket = bucket
        request.key = key

        if 'r' in kwargs:
            request.r = self._resolveNums(kwargs['r'])
        if 'pr' in kwargs:
            request.pr = self._resolveNums(kwargs['pr'])
        if 'basic_quorum' in kwargs:
            request.basic_quorum = kwargs['basic_quorum']
        if 'notfound_ok' in kwargs:
            request.notfound_ok = kwargs['notfound_ok']
        if 'if_modified' in kwargs:
            request.if_modified = kwargs['if_modified']
        if 'head' in kwargs:
            request.head = kwargs['head']
        if 'deletedvclock' in kwargs:
            request.deletedvclock = kwargs['deletedvclock']

        return self.__send(code, request)

    def get_index(self, bucket, index, startkey, endkey=None):
        code = pack('B', MSG_CODE_INDEX_REQ)
        req = RpbIndexReq(bucket=bucket, index=index)
        if endkey:
            req.qtype = RpbIndexReq.range
            req.range_min = str(startkey)
            req.range_max = str(endkey)
        else:
            req.qtype = RpbIndexReq.eq
            req.key = str(startkey)

        d = self.__send(code, req)
        d.addCallback(lambda resp: resp.keys)
        return d

    def put_new(self, bucket, key, content, vclock=None, **kwargs):
        return put(bucket, key, content, vclock, kwargs)

    def put(self, bucket, key, content, vclock=None, **kwargs):
        code = pack('B', MSG_CODE_PUT_REQ)
        request = RpbPutReq()
        request.bucket = bucket
        request.key = key

        if isinstance(content, str):
            request.content.value = content
        else:
            # assume its a dict
            request.content.value = content['value']  # mandatory

            if 'content_type' in content:
                request.content.content_type = content['content_type']
            if 'charset' in content:
                request.content.charset = content['charset']
            if 'content_encoding' in content:
                request.content.content_encoding = content['content_encoding']
            if 'vtag' in content:
                request.content.vtag = content['vtag']
            if 'last_mod' in content:
                request.content.last_mod = content['last_mod']
            if 'last_mod_usecs' in content:
                request.content.last_mod_usecs = content['last_mod_usecs']
            if 'deleted' in content:
                request.content.deleted = content['deleted']

            # write links if there are any
            if 'links' in content and isinstance(content['links'], list):
                for l in content['links']:
                    link = request.content.links.add()
                    link.bucket, link.key, link.tag = l

            # usermeta
            if 'usermeta' in content and isinstance(content['usermeta'], list):
                for l in content['usermeta']:
                    usermeta = request.content.usermeta.add()
                    usermeta.key, usermeta.value = l

            # indexes
            if 'indexes' in content and isinstance(content['indexes'], list):
                for l in content['indexes']:
                    indexes = request.content.indexes.add()
                    indexes.key, indexes.value = l

        for i in ['w', 'dw', 'pw']:
            if i in kwargs:
                setattr(request, i, self._resolveNums(kwargs[i]))

        params = [
            'if_modified', 'if_not_modified', 'if_none_match',
            'return_head', 'return_body'
        ]

        for i in params:
            if i in kwargs:
                setattr(request, i, kwargs[i])

        if vclock:
            request.vclock = vclock

        return self.__send(code, request)

    def delete(self, bucket, key, **kwargs):
        code = pack('B', MSG_CODE_DEL_REQ)
        request = RpbDelReq()
        request.bucket = bucket
        request.key = key

        if 'vclock' in kwargs and kwargs['vclock']:
            request.vclock = kwargs['vclock']

        for prop in ['rw', 'r', 'w', 'pr', 'pw', 'dw']:
            if prop in kwargs:
                setattr(request, prop, self._resolveNums(kwargs[prop]))

        return self.__send(code, request)

    # ------------------------------------------------------------------
    # Bucket Operations .. getKeys, getBuckets, get/set Bucket properties
    # ------------------------------------------------------------------
    def getKeys(self, bucket):
        """
        operates different than the other messages, as it returns more than
        one respone .. see stringReceived() for handling
        """
        code = pack('B', MSG_CODE_LIST_KEYS_REQ)
        request = RpbListKeysReq()
        request.bucket = bucket
        self.__keyList = []
        return self.__send(code, request)

    def getBuckets(self):
        """
        operates different than the other messages, as it returns more than
        one respone .. see stringReceived() for handling
        """
        code = pack('B', MSG_CODE_LIST_BUCKETS_REQ)
        return self.__send(code)

    def getBucketProperties(self, bucket):
        code = pack('B', MSG_CODE_GET_BUCKET_REQ)
        request = RpbGetBucketReq()
        request.bucket = bucket
        return self.__send(code + request.SerializeToString())

    def setBucketProperties(self, bucket, **kwargs):
        code = pack('B', MSG_CODE_SET_BUCKET_REQ)
        request = RpbSetBucketReq()
        request.bucket = bucket

        for k, v in kwargs.items():
            if k in ['n_val', 'allow_mult']:
                setattr(request.props, k, v)
            else:
                raise exceptions.RiakPBCException(
                    "Property not implemented: %s" % k)

        return self.__send(code, request)

    # ------------------------------------------------------------------
    # helper functions, message parser
    # ------------------------------------------------------------------
    def connectionMade(self):
        """
        return the protocol instance to the factory so it
        can be used directly
        """
        self.factory.connected.callback(self)

    def connectionLost(self, reason):
        self.disconnected = True

    def setTimeout(self, t):
        self.timeout = t

    def __send(self, code, request=None):
        """
        helper method for logging, sending and returning the deferred
        """
        if self.debug:
            print "[%s] %s %s" % (
                self.__class__.__name__,
                request.__class__.__name__,
                str(request).replace('\n', ' ')
            )
        if request:
            msg = code + request.SerializeToString()
        else:
            msg = code
        self.sendString(msg)
        self.factory.d = Deferred()
        if self.timeout:
            self.timeoutd = reactor.callLater(self.timeout,
                                              self._triggerTimeout)

        return self.factory.d

    def _triggerTimeout(self):
        if not self.factory.d.called:
            try:
                self.factory.d.errback(exceptions.RequestTimeout('timeout'))
            except Exception, e:
                print "Unable to handle Timeout: %s" % e

    def stringReceived(self, data):
        """
        messages contain as first byte a message type code that is used
        to map to the correct message type in self.riakResponses

        messages that dont have a body to parse return True, those are
        listed in self.nonMessages
        """
        if self.timeoutd and self.timeoutd.active():
            self.timeoutd.cancel()  # stop timeout from beeing raised

        def returnOrRaiseException(msg):
            exc = exceptions.RiakPBCException(msg)
            if self.factory.d.called:
                raise exc
            else:
                self.factory.d.errback(Failure(exc))

        # decode messagetype
        code = unpack('B', data[:1])[0]
        if self.debug:
            print "[%s] stringReceived code %s" % (self.__class__.__name__,
                self.PBMessageTypes[code])

        if code not in self.riakResponses and code not in self.nonMessages:
            returnOrRaiseException('unknown messagetype: %d' % code)

        elif code in self.nonMessages:
            # for instance ping doesnt have a message, so we just return True
            if self.debug:
                print "[%s] stringReceived empty message type %s" % (
                    self.__class__.__name__, self.PBMessageTypes[code])
            if not self.factory.d.called:
                self.factory.d.callback(True)
            return

        elif code == MSG_CODE_LIST_KEYS_RESP:
            # listKeys is special as it returns multiple response messages
            # each message can contain multiple keys
            # the last message contains a optional field "done"
            # so collect all the messages until the last one, then call the
            # callback
            response = RpbListKeysResp()
            response.ParseFromString(data[1:])
            if self.debug:
                print "[%s] %s %s" % (
                        self.__class__.__name__,
                        response.__class__.__name__,
                        str(response).replace('\n', ' ')
                    )

            self.__keyList.extend([x for x in response.keys])
            if response.HasField('done') and response.done:
                if not self.factory.d.called:
                    self.factory.d.callback(self.__keyList)
                    self.__keyList = []

        else:
            # normal handling, pick the message code, call ParseFromString()
            # on it, and return the message
            response = self.riakResponses[code]()
            if len(data) > 1:
                # if there's data, parse it, otherwise return empty object
                response.ParseFromString(data[1:])
                if self.debug:
                    print "[%s] %s %s" % (
                        self.__class__.__name__,
                        response.__class__.__name__,
                        str(response).replace('\n', ' ')
                    )

                if code == MSG_CODE_ERROR_RESP:
                    returnOrRaiseException('%s (%d)' % (
                        response.errmsg, response.errcode)
                    )

            if not self.factory.d.called:
                self.factory.d.callback(response)

    def _resolveNums(self, val):
        if isinstance(val, str):
            val = val.lower()
            if val in self.rwNums:
                return self.rwNums[val]
            else:
                raise exceptions.RiakPBCException('invalid value %s' % (val))
        else:
            return val

    def isDisconnected(self):
        return self.disconnected

    @defer.inlineCallbacks
    def quit(self):
        yield self.transport.loseConnection()


class RiakPBCClientFactory(ClientFactory):
    protocol = RiakPBC
    noisy = False

    def __init__(self):
        self.d = Deferred()
        self.connected = Deferred()

    def clientConnectionFailed(self, connector, reason):
        self.connected.errback(reason)


class RiakPBCClient(object):
    def connect(self, host, port):
        factory = RiakPBCClientFactory()
        reactor.connectTCP(host, port, factory)
        return factory.connected



"""
.. module:: client.py

RiakClient class

"""

import random
import base64
import urllib
import json
from twisted.internet import defer

from riakasaurus import mapreduce, bucket
from riakasaurus.search import RiakSearch

from riakasaurus import transport


class RiakClient(object):
    """
    The RiakClient object holds information necessary to connect to
    Riak.
    """
    def __init__(self, host='127.0.0.1', port=8098,
                prefix='riak', mapred_prefix='mapred',
                client_id=None, r_value="default", w_value="default",
                dw_value="default", transport=transport.HTTPTransport,
                request_timeout=None):
        """
        Construct a new RiakClient object.

        If a client_id is not provided, generate a random one.
        """
        self._host = host
        self._port = port
        self._prefix = prefix
        self._mapred_prefix = mapred_prefix
        if client_id:
            self._client_id = client_id
        else:
            self._client_id = 'py_' + base64.b64encode(
                                      str(random.randint(1, 1073741824)))
        self._r = r_value
        self._w = w_value
        self._dw = dw_value

        self._rw = "default"
        self._pr = "default"
        self._pw = "default"

        self._encoders = {'application/json': json.dumps,
                          'text/json': json.dumps}
        self._decoders = {'application/json': json.loads,
                          'text/json': json.loads}
        self._solr = None

        self.request_timeout = request_timeout

        self.transport = transport(self)

    def setRequestTimeout(self, timeout):
        self.request_timeout = timeout

    def get_transport(self):
        return self.transport

    def get_r(self):
        """
        Get the R-value setting for this RiakClient. (default 2)
        :returns: integer representing current r-value
        .. todo:: remove accessor
        """
        return self._r

    def get_rw(self):
        """
        Get the RW-value for this ``RiakClient`` instance. (default "quorum")

        :rtype: integer
        """
        return self._rw

    def set_rw(self, rw):
        """
        Set the RW-value for this ``RiakClient`` instance. See :func:`set_r`
        for a description of how these values are used.

        :param rw: The RW value.
        :type rw: integer
        :rtype: self
        """
        self._rw = rw
        return self

    def get_pw(self):
        """
        Get the PW-value setting for this ``RiakClient``. (default 0)

        :rtype: integer
        """
        return self._pw

    def set_pw(self, pw):
        """
        Set the PW-value for this ``RiakClient`` instance. See :func:`set_r`
        for a description of how these values are used.

        :param pw: The W value.
        :type pw: integer
        :rtype: self
        """
        self._pw = pw
        return self

    def get_pr(self):
        """
        Get the PR-value setting for this ``RiakClient``. (default 0)

        :rtype: integer
        """
        return self._pr

    def set_pr(self, pr):
        """
        Set the PR-value for this ``RiakClient`` instance. See :func:`set_r`
        for a description of how these values are used.

        :param pr: The PR value.
        :type pr: integer
        :rtype: self
        """
        self._pr = pr
        return self

    def set_r(self, r):
        """
        Set the R-value for this RiakClient. This value will be used
        for any calls to get(...) or get_binary(...) where where 1) no
        R-value is specified in the method call and 2) no R-value has
        been set in the RiakBucket.
        @param integer r - The R value.
        @return self
        .. todo:: remove accessor
        """
        self._r = r
        return self

    def get_w(self):
        """
        Get the W-value setting for this RiakClient. (default 2)
        @return integer
        .. todo:: remove accessor
        """
        return self._w

    def set_w(self, w):
        """
        Set the W-value for this RiakClient. See set_r(...) for a
        description of how these values are used.
        @param integer w - The W value.
        @return self
        .. todo:: remove accessor
        """
        self._w = w
        return self

    def get_dw(self):
        """
        Get the DW-value for this ClientOBject. (default 2)
        @return integer
        .. todo:: remove accessor
        """
        return self._dw

    def set_dw(self, dw):
        """
        Set the DW-value for this RiakClient. See set_r(...) for a
        description of how these values are used.
        @param integer dw - The DW value.
        @return self
        .. todo:: remove accessor
        """
        self._dw = dw
        return self

    def get_client_id(self):
        """
        Get the client_id for this RiakClient.
        @return string
        .. todo:: remove accessor
        """
        return self._client_id

    def set_client_id(self, client_id):
        """
        Set the client_id for this RiakClient. Should not be called
        unless you know what you are doing.
        @param string client_id - The new client_id.
        @return self
        .. todo:: remove accessor
        """
        self._client_id = client_id
        return self

    def get_encoder(self, content_type):
        """
        Get the encoding function for the provided content type.
        """
        if content_type in self._encoders:
            return self._encoders[content_type]

    def set_encoder(self, content_type, encoder):
        """
        Set the encoding function for the provided content type.

        :param encoder:
        :type encoder: function
        """
        self._encoders[content_type] = encoder
        return self

    def get_decoder(self, content_type):
        """
        Get the decoding function for the provided content type.
        """
        if content_type in self._decoders:
            return self._decoders[content_type]

    def set_decoder(self, content_type, decoder):
        """
        Set the decoding function for the provided content type.

        :param decoder:
        :type decoder: function
        """
        self._decoders[content_type] = decoder
        return self

    def bucket(self, name):
        """
        Get the bucket by the specified name. Since buckets always exist,
        this will always return a RiakBucket.
        :returns: RiakBucket instance.
        """
        return bucket.RiakBucket(self, name)

    def is_alive(self):
        """
        Check if the Riak server for this RiakClient is alive.
        :returns: True if alive -- via deferred.
        """

        return self.transport.ping()

    def add(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.add()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self)
        return mr.add(*args)

    def search(self, *args):
        """
        Start assembling a Map/Reduce operation for Riak Search
        see RiakMapReduce.search()
        """
        mr = mapreduce.RiakMapReduce(self)
        return mr.search(*args)

    def link(self, args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.link()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self)
        return mr.link(*args)

    def list_buckets(self):
        """
        Retrieve a list of all buckets.

        :returns: list -- via deferred
        """

        return self.transport.get_buckets()

    def index(self, *args):
        """
        Start assembling a Map/Reduce operation based on secondary
        index query results.

        :rtype: :class:`RiakMapReduce`
        """
        mr = mapreduce.RiakMapReduce(self)
        return mr.index(*args)

    def map(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.map()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self)
        return mr.map(*args)

    def reduce(self, *args):
        """
        Start assembling a Map/Reduce operation.
        see RiakMapReduce.reduce()
        :returns: RiakMapReduce
        """
        mr = mapreduce.RiakMapReduce(self)
        return mr.reduce(*args)

    def solr(self):
        if self._solr is None:
            self._solr = RiakSearch(self, host=self._host, port=self._port)

        return self._solr


if __name__ == "__main__":
    pass



#!/usr/bin/env python
"""
riakasaurus trial test file.
riakasaurus _must_ be on your PYTHONPATH

"""

import json
import random
from twisted.trial import unittest
from twisted.python import log
from twisted.internet import defer

VERBOSE = False

from riakasaurus import riak

# uncomment to activate logging
# import sys
# log.startLogging(sys.stderr)

RIAK_CLIENT_ID = 'TEST'
BUCKET_PREFIX = 'riakasaurus.tests.'

JAVASCRIPT_SUM = """
function(v) {
  x = v.reduce(function(a,b){ return a + b }, 0);
  return [x];
}
"""


def randint():
    """Generate nice random int for our test."""
    return random.randint(1, 999999)


class Tests(unittest.TestCase):
    """
    trial unit tests.
    """

    test_keys = ['foo', 'foo1', 'foo2', 'foo3', 'bar', 'baz', 'ba_foo1',
                 'blue_foo1']

    @defer.inlineCallbacks
    def setUp(self):
        self.client = riak.RiakClient(client_id=RIAK_CLIENT_ID)
        self.bucket_name = BUCKET_PREFIX + self.id().rsplit('.', 1)[-1]
        self.bucket = self.client.bucket(self.bucket_name)
        yield self.bucket.enable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def tearDown(self):
        yield self.bucket.disable_search()
        yield self.bucket.purge_keys()

    @defer.inlineCallbacks
    def test_javascript_source_map(self):
        """javascript mapping"""
        log.msg('*** javascript_source_map')
        # Create the object...
        obj = self.bucket.new("foo", 2)
        yield obj.store()
        # Run the map...
        job = self.client \
                .add(self.bucket_name, "foo") \
                .map("function (v) { return [JSON.parse(v.values[0].data)]; }")
        result = yield job.run()
        self.assertEqual(result, [2])
        log.msg('done javascript_source_map')

    @defer.inlineCallbacks
    def test_javascript_named_map(self):
        """javascript mapping with named map"""
        log.msg('*** javascript_named_map')
        # Create the object...
        obj = self.bucket.new("foo", 2)
        yield obj.store()
        # Run the map...
        job = self.client \
                .add(self.bucket_name, "foo") \
                .map("Riak.mapValuesJson")
        result = yield job.run()
        self.assertEqual(result, [2])
        log.msg('done javascript_named_map')

    @defer.inlineCallbacks
    def test_javascript_source_map_reduce(self):
        """javascript map reduce"""
        log.msg('*** javascript_source_map_reduce')
        # Create the object...
        yield self.bucket.new("foo", 2).store()
        yield self.bucket.new("bar", 3).store()
        yield self.bucket.new("baz", 4).store()
        # Run the map...
        job = self.client \
                .add(self.bucket_name, "foo") \
                .add(self.bucket_name, "bar") \
                .add(self.bucket_name, "baz") \
                .map("function (v) { return [1]; }") \
                .reduce(JAVASCRIPT_SUM)
        result = yield job.run()
        self.assertEqual(result, [3])
        log.msg('done javascript_source_map_reduce')

    @defer.inlineCallbacks
    def test_javascript_named_map_reduce(self):
        """javascript map reduce by name"""
        log.msg('*** javascript_named_map_reduce')
        # Create the object...
        yield self.bucket.new("foo", 2).store()
        yield self.bucket.new("bar", 3).store()
        yield self.bucket.new("baz", 4).store()
        # Run the map...
        job = self.client \
                .add(self.bucket_name, "foo") \
                .add(self.bucket_name, "bar") \
                .add(self.bucket_name, "baz") \
                .map("Riak.mapValuesJson") \
                .reduce("Riak.reduceSum")
        result = yield job.run()
        self.assertEqual(result, [9])
        log.msg('done javascript_named_map_reduce')

    @defer.inlineCallbacks
    def test_javascript_key_filter_map_reduce(self):
        """javascript map/reduce using key filters"""
        log.msg("javascript map reduce with key filter")
        yield self.bucket.new("foo", 2).store()
        yield self.bucket.new("bar", 3).store()
        yield self.bucket.new("baz", 4).store()
        # Run the map...
        job = self.client \
                .add({"bucket": self.bucket_name,
                      "key_filters": [["starts_with", "ba"]]}) \
                .map("function (v) { return [1]; }") \
                .reduce(JAVASCRIPT_SUM)
        result = yield job.run()
        self.assertEqual(result, [2])
        log.msg('done javascript_key_filter_map_reduce')

    def test_javascript_bucket_map_reduce(self):
        """javascript bucket map reduce"""
        log.msg('*** javascript_bucket_map_reduce')
        # Create the object...
        yield self.bucket.new("foo", 2).store()
        yield self.bucket.new("bar", 3).store()
        yield self.bucket.new("baz", 4).store()
        # Run the map...
        job = self.client \
                .add(self.bucket_name) \
                .map("Riak.mapValuesJson") \
                .reduce("Riak.reduceSum")
        result = yield job.run()
        self.assertEqual(result, [9])

    @defer.inlineCallbacks
    def test_javascript_arg_map_reduce(self):
        """javascript arguments map reduce"""
        log.msg('*** javascript_arg_map_reduce')
        # Create the object...
        obj = self.bucket.new("foo", 2)
        yield obj.store()
        # Run the map...
        job = self.client \
                .add(self.bucket_name, "foo", 5) \
                .add(self.bucket_name, "foo", 10) \
                .add(self.bucket_name, "foo", 15) \
                .add(self.bucket_name, "foo", -15) \
                .add(self.bucket_name, "foo", -5) \
                .map("function(v, arg) { return [arg]; }") \
                .reduce("Riak.reduceSum")
        result = yield job.run()
        self.assertEqual(result, [10])
        log.msg('done javascript_arg_map_reduce')
